# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from collections.abc import Iterable

from gitfourchette import settings
from gitfourchette import tasks
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.tasks import TaskBook
from gitfourchette.toolbox import *


class MainToolBar(QToolBar):
    openDialog = Signal()
    openPrefs = Signal()
    reveal = Signal()
    openTerminal = Signal()
    pull = Signal()
    push = Signal()

    backAction: QAction
    forwardAction: QAction
    terminalAction: QAction
    recentAction: QAction

    def __init__(self, parent: QWidget):
        super().__init__(englishTitleCase(_("Show toolbar")), parent)

        self.setObjectName("GFToolbar")
        self.setMovable(False)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onCustomContextMenuRequested)

        self.visibilityChanged.connect(self.onVisibilityChanged)
        self.toolButtonStyleChanged.connect(self.onToolButtonStyleChanged)
        self.iconSizeChanged.connect(self.onIconSizeChanged)

        self.backAction = TaskBook.toolbarAction(self, tasks.JumpBack).toQAction(self)
        self.forwardAction = TaskBook.toolbarAction(self, tasks.JumpForward).toQAction(self)

        self.workdirAction = TaskBook.toolbarAction(self, tasks.JumpToUncommittedChanges).toQAction(self)
        self.headAction = TaskBook.toolbarAction(self, tasks.JumpToHEAD).toQAction(self)

        self.terminalAction = ActionDef(
            _("Terminal"), self.openTerminal, icon="terminal",
            shortcuts=GlobalShortcuts.openTerminal,
            tip=_("Open a terminal in the repo")
        ).toQAction(self)

        self.recentAction = ActionDef(
            _("Open…"), self.openDialog, icon="git-folder",
            shortcuts=QKeySequence.StandardKey.Open,
            tip=_("Open a Git repo on your machine")
        ).toQAction(self)

        self.settingsAction = ActionDef(
            _("Settings"), self.openPrefs, icon="git-settings",
            shortcuts=QKeySequence.StandardKey.Preferences,
            tip=_("Configure {app}", app=qAppName())
        ).toQAction(self)

        defs = [
            self.backAction,
            self.forwardAction,
            self.workdirAction,
            self.headAction,
            ActionDef.SEPARATOR,

            TaskBook.toolbarAction(self, tasks.NewStash),
            TaskBook.toolbarAction(self, tasks.NewBranchFromHead),
            ActionDef.SEPARATOR,
            TaskBook.toolbarAction(self, tasks.FetchRemotes),
            TaskBook.toolbarAction(self, tasks.PullBranch),
            TaskBook.toolbarAction(self, tasks.PushBranch),
            ActionDef.SPACER,

            self.terminalAction,
            ActionDef(_("Reveal"), self.reveal, icon="reveal",
                      shortcuts=GlobalShortcuts.openRepoFolder,
                      tip=_("Open repo folder in file manager")),

            self.recentAction,

            ActionDef.SEPARATOR,
            self.settingsAction,
        ]
        ActionDef.addToQToolBar(self, *defs)

        self.recentAction.setIconVisibleInMenu(True)
        recentButton = self.widgetForAction(self.recentAction)
        assert isinstance(recentButton, QToolButton)
        recentButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)

        self.setToolButtonStyle(settings.prefs.toolBarButtonStyle)
        self.setIconSize(QSize(settings.prefs.toolBarIconSize, settings.prefs.toolBarIconSize))

        self.updateNavButtons()

    def setToolButtonStyle(self, style: Qt.ToolButtonStyle):
        # Resolve style
        if style == Qt.ToolButtonStyle.ToolButtonFollowStyle:
            styleHint = QApplication.style().styleHint(QStyle.StyleHint.SH_ToolButtonStyle)
            style = Qt.ToolButtonStyle(styleHint)

        super().setToolButtonStyle(style)

        # Hide back/forward button text with ToolButtonTextBesideIcon
        if style != Qt.ToolButtonStyle.ToolButtonTextOnly:
            for navAction in (self.backAction, self.forwardAction):
                navButton = self.widgetForAction(navAction)
                assert isinstance(navButton, QToolButton)
                navButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

        if style == Qt.ToolButtonStyle.ToolButtonTextBesideIcon:
            for navAction in (self.headAction, self.workdirAction, self.settingsAction):
                navButton = self.widgetForAction(navAction)
                assert isinstance(navButton, QToolButton)
                navButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

    def onCustomContextMenuRequested(self, localPoint: QPoint):
        globalPoint = self.mapToGlobal(localPoint)

        def textPositionAction(name, style):
            return ActionDef(name,
                             callback=lambda: self.setToolButtonStyle(style),
                             checkState=1 if self.toolButtonStyle() == style else -1,
                             radioGroup="TextPosition")

        def iconSizeAction(name, size):
            return ActionDef(name,
                             callback=lambda: self.setIconSize(QSize(size, size)),
                             checkState=1 if self.iconSize().width() == size else -1,
                             radioGroup="IconSize")

        menu = ActionDef.makeQMenu(self, [
            self.toggleViewAction(),

            ActionDef.SEPARATOR,

            ActionDef(
                _("Text Position"),
                submenu=[
                    textPositionAction(_("Icons Only"), Qt.ToolButtonStyle.ToolButtonIconOnly),
                    textPositionAction(_("Text Only"), Qt.ToolButtonStyle.ToolButtonTextOnly),
                    textPositionAction(_("Text Alongside Icons"), Qt.ToolButtonStyle.ToolButtonTextBesideIcon),
                    textPositionAction(_("Text Under Icons"), Qt.ToolButtonStyle.ToolButtonTextUnderIcon),
                ]
            ),

            ActionDef(
                _("Icon Size"),
                submenu=[
                    iconSizeAction(_("Small"), 16),
                    iconSizeAction(_("Medium"), 20),
                    iconSizeAction(_("Large"), 24),
                    iconSizeAction(_("Huge"), 32),
                ]
            ),
        ])

        menu.popup(globalPoint)

    def onVisibilityChanged(self, visible: bool):
        # self.window().setUnifiedTitleAndToolBarOnMac(visible)
        if visible == settings.prefs.showToolBar:
            return
        settings.prefs.showToolBar = visible
        settings.prefs.setDirty()

    def onToolButtonStyleChanged(self, style: Qt.ToolButtonStyle):
        if style == settings.prefs.toolBarButtonStyle:
            return
        settings.prefs.toolBarButtonStyle = style
        settings.prefs.setDirty()

    def onIconSizeChanged(self, size: QSize):
        w = size.width()
        if w == settings.prefs.toolBarIconSize:
            return
        settings.prefs.toolBarIconSize = w
        settings.prefs.setDirty()

    def updateNavButtons(self, back: bool = False, forward: bool = False):
        self.backAction.setEnabled(back)
        self.forwardAction.setEnabled(forward)

    def setTerminalActions(self, newActions: Iterable[QAction]):
        terminalAction = self.terminalAction
        terminalButton = self.widgetForAction(terminalAction)
        terminalMenu = terminalAction.menu()

        assert isinstance(terminalButton, QToolButton)

        if newActions:
            if not terminalMenu:
                terminalMenu = QMenu(terminalButton)
                terminalMenu.setObjectName("MainToolBarTerminalMenu")
                terminalAction.setMenu(terminalMenu)
                terminalButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
            terminalMenu.clear()
            terminalMenu.addActions(newActions)
        elif terminalMenu:
            # The QToolButton will show an arrow as long as there's a menu, even if it's empty
            terminalMenu.deleteLater()
            terminalAction.setMenu(None)
            terminalButton.setPopupMode(QToolButton.ToolButtonPopupMode.DelayedPopup)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import typing
from typing import Literal

from gitfourchette.application import GFApplication
from gitfourchette.diffbuttons import DiffButtons
from gitfourchette.diffview.diffview import DiffView
from gitfourchette.diffview.specialdiffview import SpecialDiffView
from gitfourchette.filelists.committedfiles import CommittedFiles
from gitfourchette.filelists.dirtyfiles import DirtyFiles
from gitfourchette.filelists.filelist import FileList
from gitfourchette.filelists.stagedfiles import StagedFiles
from gitfourchette.forms.banner import Banner
from gitfourchette.forms.conflictview import ConflictView
from gitfourchette.forms.contextheader import ContextHeader
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.nav import NavContext, NavLocator, NavFlags
from gitfourchette.qt import *
from gitfourchette.tasks import TaskBook, AmendCommit, NewCommit, NewStash
from gitfourchette.toolbox import *

FileStackPage = Literal["workdir", "commit"]
DiffStackPage = Literal["text", "special", "conflict"]

FILEHEADER_HEIGHT = 24

logger = logging.getLogger(__name__)


def gridPadding():
    return QSpacerItem(3, 1, QSizePolicy.Policy.Fixed)


class DiffArea(QWidget):
    def __init__(self, repoModel, parent):
        super().__init__(parent)
        self.setObjectName("CommitExplorer")

        fileStack = self._makeFileStack(repoModel)
        diffContainer = self._makeDiffContainer(repoModel)

        splitter = QSplitter(Qt.Orientation.Horizontal, self)
        splitter.setObjectName("Split_DiffArea")

        contextHeader = ContextHeader(self)

        diffBanner = Banner(self, orientation=Qt.Orientation.Horizontal)
        diffBanner.setProperty("class", "diff")
        diffBanner.setVisible(False)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(QMargins())
        layout.setSpacing(0)
        layout.addWidget(contextHeader)
        layout.addWidget(diffBanner)
        layout.addWidget(QFaintSeparator(self))
        layout.addWidget(splitter, 1)

        splitter.addWidget(fileStack)
        splitter.addWidget(diffContainer)
        splitter.setSizes([260, 500])
        splitter.setStretchFactor(0, 0)  # don't auto-stretch file lists when resizing window
        splitter.setStretchFactor(1, 1)
        splitter.setChildrenCollapsible(False)

        self.fileStack = fileStack
        self.diffBanner = diffBanner
        self.contextHeader = contextHeader

        for passiveWidget in (
                self.diffHeader,
                self.committedHeader,
                self.dirtyHeader,
                self.stagedHeader
        ):
            passiveWidget.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)

        GFApplication.instance().prefsChanged.connect(self.diffButtons.refreshPrefs)
        self.diffButtons.refreshPrefs()

        # Ignore height in size policy to keep DiffArea from jumping around when we're showing a banner.
        self.setSizePolicy(self.sizePolicy().horizontalPolicy(), QSizePolicy.Policy.Ignored)
        self.setMinimumHeight(175)

    # -------------------------------------------------------------------------
    # Constructor helpers

    def _makeFileStack(self, repoModel):
        dirtyContainer = self._makeDirtyContainer(repoModel)
        stageContainer = self._makeStageContainer(repoModel)
        committedFilesContainer = self._makeCommittedFilesContainer(repoModel)

        stagingSplitter = QSplitter(Qt.Orientation.Vertical, self)
        stagingSplitter.addWidget(dirtyContainer)
        stagingSplitter.addWidget(stageContainer)
        stagingSplitter.setObjectName("Split_Staging")
        stagingSplitter.setChildrenCollapsible(False)

        fileStack = QStackedWidget()
        fileStack.addWidget(stagingSplitter)
        fileStack.addWidget(committedFilesContainer)
        return fileStack

    def _makeDirtyContainer(self, repoModel):
        header = QElidedLabel(" ")
        header.setProperty("class", "panelTitle")
        header.setObjectName("dirtyHeader")
        header.setToolTip(_("Unstaged files: will not be included in the commit unless you stage them."))
        header.setMinimumHeight(FILEHEADER_HEIGHT)
        header.setEnabled(False)

        dirtyFiles = DirtyFiles(repoModel, self)

        stageButton = QToolButton(self)
        stageButton.setObjectName("stageButton")
        stageButton.setText(_("Stage"))
        stageButton.setIcon(stockIcon("git-stage"))
        stageButton.setToolTip(_("Stage selected files"))
        appendShortcutToToolTip(stageButton, GlobalShortcuts.stageHotkeys[0])

        discardButton = QToolButton(self)
        discardButton.setObjectName("discardButton")
        discardButton.setText(_("Discard"))
        discardButton.setIcon(stockIcon("git-discard"))
        discardButton.setToolTip(_("Discard changes in selected files"))
        appendShortcutToToolTip(discardButton, GlobalShortcuts.discardHotkeys[0])

        container = QWidget(self)
        layout = QGridLayout(container)
        layout.setSpacing(0)  # automatic frameless list views on KDE Plasma 6 Breeze
        layout.setContentsMargins(QMargins())
        # Row 0
        layout.addItem(gridPadding(),           0, 0)
        layout.addWidget(header,                0, 1)
        layout.addWidget(stageButton,           0, 2)
        layout.addWidget(discardButton,         0, 3)
        # Row 1
        layout.addItem(QSpacerItem(1, 1),       1, 0, 1, 4)
        # Row 2
        layout.addWidget(dirtyFiles.searchBar,  2, 0, 1, 4)
        # Row 3
        layout.addWidget(dirtyFiles,            3, 0, 1, 4)
        layout.setRowStretch(3, 100)

        stageButton.clicked.connect(dirtyFiles.stage)
        discardButton.clicked.connect(dirtyFiles.discard)
        dirtyFiles.selectedCountChanged.connect(lambda n: stageButton.setEnabled(n > 0))
        dirtyFiles.selectedCountChanged.connect(lambda n: discardButton.setEnabled(n > 0))

        self.dirtyFiles = dirtyFiles
        self.dirtyHeader = header
        self.stageButton = stageButton
        self.discardButton = discardButton

        return container

    def _makeStageContainer(self, repoModel):
        header = QElidedLabel(" ")
        header.setObjectName("stagedHeader")
        header.setProperty("class", "panelTitle")
        header.setToolTip(_("Staged files: will be included in the commit."))
        header.setMinimumHeight(FILEHEADER_HEIGHT)
        header.setEnabled(False)

        stagedFiles = StagedFiles(repoModel, self)

        unstageButton = QToolButton(self)
        unstageButton.setObjectName("unstageButton")
        unstageButton.setText(_("Unstage"))
        unstageButton.setIcon(stockIcon("git-unstage"))
        unstageButton.setToolTip(_("Unstage selected files"))
        appendShortcutToToolTip(unstageButton, GlobalShortcuts.discardHotkeys[0])

        commitButton = QToolButton(self)
        commitButton.setObjectName("commitButton")
        commitButton.setText(_p("verb", "Commit"))
        commitButton.setIcon(stockIcon("git-commit", "gray=#599E5E"))
        commitButton.setToolTip(appendShortcutToToolTipText(TaskBook.tips[NewCommit], TaskBook.shortcuts[NewCommit][0]))
        commitButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        commitButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        commitButton.setAutoRaise(True)
        commitButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
        commitButton.setMaximumHeight(FILEHEADER_HEIGHT)
        commitButton.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)

        # Connect signals
        unstageButton.clicked.connect(stagedFiles.unstage)
        stagedFiles.selectedCountChanged.connect(lambda n: unstageButton.setEnabled(n > 0))

        commitButton.clicked.connect(lambda: NewCommit.invoke(self))
        commitButtonMenu = ActionDef.makeQMenu(
            commitButton,
            [
                TaskBook.action(self, NewCommit),
                TaskBook.action(self, AmendCommit),
                TaskBook.action(self, NewStash),
            ])
        # Prevent shortcuts from taking over
        for action in commitButtonMenu.actions():
            action.setShortcutContext(Qt.ShortcutContext.WidgetShortcut)
        commitButton.setMenu(commitButtonMenu)

        # Lay out container
        container = QWidget(self)
        layout = QGridLayout(container)
        layout.setContentsMargins(QMargins())
        layout.setSpacing(0)  # automatic frameless list views on KDE Plasma 6 Breeze
        # Row 0
        layout.addItem(gridPadding(),           0, 0)
        layout.addWidget(header,                0, 1)
        layout.addWidget(unstageButton,         0, 2)
        # Row 1
        layout.addItem(QSpacerItem(1, 1),       1, 0)
        # Row 2
        layout.addWidget(stagedFiles.searchBar, 2, 0, 1, 3)  # row col rowspan colspan
        layout.addWidget(stagedFiles,           3, 0, 1, 3)
        layout.addWidget(commitButton,          4, 0, 1, 3)
        layout.setRowStretch(3, 100)

        # Save references
        self.stagedHeader = header
        self.stagedFiles = stagedFiles
        self.unstageButton = unstageButton
        self.commitButton = commitButton

        return container

    def _makeCommittedFilesContainer(self, repoModel):
        committedFiles = CommittedFiles(repoModel, self)

        header = QElidedLabel(" ")
        header.setObjectName("committedHeader")
        header.setProperty("class", "panelTitle")
        header.setMinimumHeight(FILEHEADER_HEIGHT)
        header.setEnabled(False)

        container = QWidget(self)
        layout = QGridLayout(container)
        layout.setContentsMargins(QMargins())
        layout.setSpacing(0)  # automatic frameless list views on KDE Plasma 6 Breeze
        layout.addItem(gridPadding(),               0, 0)
        layout.addWidget(header,                    0, 1)
        layout.addItem(gridPadding(),               0, 2)
        layout.addWidget(committedFiles.searchBar,  1, 0, 1, 3)
        layout.addItem(QSpacerItem(1, 1),           2, 0, 1, 3)
        layout.addWidget(committedFiles,            3, 0, 1, 3)

        self.committedFiles = committedFiles
        self.committedHeader = header
        return container

    def _makeDiffContainer(self, repoModel):
        header = QLabel(" ")
        header.setObjectName("diffHeader")
        header.setProperty("class", "panelTitle")
        header.setMinimumHeight(FILEHEADER_HEIGHT)
        header.setContentsMargins(4, 0, 4, 0)
        header.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        # Absorb horizontal space so the path stays left and toolbar stays right.
        header.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Minimum)

        diffTools = DiffButtons(self)

        topContainer = QWidget(self)
        topLayout = QHBoxLayout(topContainer)
        topLayout.setContentsMargins(0, 0, 0, 0)
        topLayout.setSpacing(0)
        topLayout.addWidget(header, 1)
        topLayout.addWidget(diffTools, 0)

        diff = DiffView(self)

        diffViewContainer = QWidget(self)
        diffViewContainerLayout = QVBoxLayout(diffViewContainer)
        diffViewContainerLayout.setSpacing(0)
        diffViewContainerLayout.setContentsMargins(0, 0, 0, 0)
        diffViewContainerLayout.addWidget(diff.searchBar)
        diffViewContainerLayout.addWidget(diff)

        specialDiff = SpecialDiffView(self)

        conflict = ConflictView(repoModel, self)
        conflictScroll = QScrollArea()
        conflictScroll.setWidget(conflict)
        conflictScroll.setWidgetResizable(True)

        stack = QStackedWidget(self)
        # Add widgets in same order as DiffStackPage
        stack.addWidget(diffViewContainer)
        stack.addWidget(specialDiff)
        stack.addWidget(conflictScroll)
        stack.setCurrentIndex(0)

        stackContainer = QWidget(self)
        layout = QVBoxLayout(stackContainer)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        layout.addWidget(topContainer)
        layout.addWidget(stack)

        self.diffHeader = header
        self.diffStack = stack
        self.conflictView = conflict
        self.specialDiffView = specialDiff
        self.diffView = diff
        self.diffButtons = diffTools

        return stackContainer

    def applyCustomStyling(self):
        for smallButton in (
                self.discardButton,
                self.unstageButton,
                self.stageButton,
                *self.diffButtons.buttons,
        ):
            smallButton.setMaximumHeight(FILEHEADER_HEIGHT)
            smallButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            smallButton.setAutoRaise(True)
            smallButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

        for button in self.stageButton, self.unstageButton, self.discardButton:
            button.setEnabled(False)

        for button in self.stageButton, self.unstageButton:
            button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        # Smaller font for header text
        for smallWidget in (
                self.contextHeader,
                self.diffHeader,
                self.committedHeader,
                self.dirtyHeader,
                self.stagedHeader,
                self.stageButton,
                self.unstageButton,
                self.discardButton,
                *self.diffButtons.buttons,
        ):
            tweakWidgetFont(smallWidget, 90)

    # -------------------------------------------------------------------------
    # File navigation

    def selectNextFile(self, down=True):
        page = self.fileStackPage()
        widgets: list[FileList]
        if page == "commit":
            widgets = [self.committedFiles]
        elif page == "workdir":
            widgets = [self.dirtyFiles, self.stagedFiles]
        else:
            raise NotImplementedError(f"Unknown FileStackPage {page}")

        numWidgets = len(widgets)
        selections = [w.selectedIndexes() for w in widgets]
        lengths = [w.model().rowCount() for w in widgets]

        # find widget to start from: topmost widget that has any selection
        leader = -1
        for i, selection in enumerate(selections):
            if selection:
                leader = i
                break

        if leader < 0:
            # selection empty; pick first non-empty widget as leader
            leader = 0
            row = 0
            while (leader < numWidgets) and (lengths[leader] == 0):
                leader += 1
        else:
            # get selected row in leader widget - TODO: this may not be accurate when multiple rows are selected
            row = selections[leader][-1].row()

            if down:
                row += 1
                while (leader < numWidgets) and (row >= lengths[leader]):
                    # out of rows in leader widget; jump to first row in next widget
                    leader += 1
                    row = 0
            else:
                row -= 1
                while (leader >= 0) and (row < 0):
                    # out of rows in leader widget; jump to last row in prev widget
                    leader -= 1
                    if leader >= 0:
                        row = lengths[leader] - 1

        # if we have a new valid selection, apply it, otherwise bail
        if 0 <= leader < numWidgets and 0 <= row < lengths[leader]:
            widgets[leader].setFocus()
            with QSignalBlockerContext(widgets[leader]):
                widgets[leader].clearSelection()
            widgets[leader].selectRow(row)
        else:
            # No valid selection
            QApplication.beep()

            # Focus on the widget that has some selected files in it
            for w in widgets:
                if len(w.selectedIndexes()) > 0:
                    w.setFocus()
                    break

    def fileListByContext(self, context: NavContext) -> FileList:
        if context == NavContext.STAGED:
            return self.stagedFiles
        elif context == NavContext.UNSTAGED:
            return self.dirtyFiles
        else:
            return self.committedFiles

    def setUpForLocator(self, locator: NavLocator) -> NavLocator:
        """
        Show relevant FileList widget, select correct file in it,
        and adjust auxiliary widgets (stage/unstage/discard buttons).

        If the desired path isn't available in the FileList,
        returns a new locator with a blank path.
        """
        fileList = self.fileListByContext(locator.context)

        with QSignalBlockerContext(self.dirtyFiles, self.stagedFiles, self.committedFiles):
            # Select correct row in FileList
            hasFile = False
            if locator.path:
                # Fix multiple "ghost" selections in DirtyFiles/StagedFiles with JumpBackOrForward.
                if not locator.hasFlags(NavFlags.BypassFileSelect):
                    fileList.clearSelection()
                # Select the file, if possible
                hasFile = fileList.selectFile(locator.path)

            # Blank selection?
            if not hasFile and locator.context != NavContext.SPECIAL:
                locator = locator.replace(path="")
                fileList.clearSelection()

            # Special treatment for workdir
            if locator.context.isWorkdir():
                staged = locator.context == NavContext.STAGED

                # Sync workdir buttons
                self.stageButton.setEnabled(hasFile and not staged)
                self.discardButton.setEnabled(hasFile and not staged)
                self.unstageButton.setEnabled(hasFile and staged)

                # Clear selection in opposite FileList
                oppositeFileList = self.dirtyFiles if staged else self.stagedFiles
                oppositeFileList.clearSelection()
                if hasFile:
                    oppositeFileList.highlightCounterpart(locator)

            # Set correct card in fileStack (after selecting the file to avoid flashing)
            self.setFileStackPageByContext(locator.context)

        return locator

    # -------------------------------------------------------------------------
    # Clear

    def clearDocument(self, locator: NavLocator):
        # Enter empty special page
        self.specialDiffView.clear()
        self.setDiffStackPage("special")

        # Might as well free up any memory taken by DiffView document
        self.diffView.clear()

        self.diffHeader.setText(" ")

        self.setUpForLocator(locator)

    # -------------------------------------------------------------------------
    # Stacked widget helpers

    @property
    def _fileStackPageValues(self):
        return typing.get_args(FileStackPage)

    def fileStackPage(self) -> FileStackPage:
        return self._fileStackPageValues[self.fileStack.currentIndex()]

    def setFileStackPage(self, p: FileStackPage):
        self.fileStack.setCurrentIndex(self._fileStackPageValues.index(p))

    def setFileStackPageByContext(self, context: NavContext):
        page: FileStackPage = "workdir" if context.isWorkdir() else "commit"
        self.setFileStackPage(page)

    @property
    def _diffStackPageValues(self):
        return typing.get_args(DiffStackPage)

    def diffStackPage(self) -> DiffStackPage:
        return self._diffStackPageValues[self.diffStack.currentIndex()]

    def setDiffStackPage(self, p: DiffStackPage):
        self.diffStack.setCurrentIndex(self._diffStackPageValues.index(p))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import html
import logging
import sys
import traceback
from collections.abc import Callable
from typing import Literal

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox.excutils import shortenTracebackPath
from gitfourchette.toolbox.qtutils import onAppThread, MakeNonNativeDialog
from gitfourchette.toolbox.textutils import stripHtml, ulify

logger = logging.getLogger(__name__)

MessageBoxIconName = Literal['warning', 'information', 'question', 'critical']

_excMessageBoxQueue: list[QMessageBox] = []


def excMessageBox(
        exc,
        title="",
        message="",
        parent=None,
        printExc=True,
        showExcSummary=True,
        icon: MessageBoxIconName = 'critical',
        abortUnitTest=True,
):
    try:
        if exc.__class__.__name__ == 'NoRepoWidgetError':
            title = _("No repository")
            message = _("Please open a repository before performing this action.")
            showExcSummary = False
            icon = 'information'

        isCritical = icon == 'critical'

        if printExc:
            traceback.print_exception(exc.__class__, exc, exc.__traceback__)

        # Without a parent, show() won't work. Try to find a QMainWindow to use as the parent.
        if not parent:
            for tlw in QApplication.topLevelWidgets():
                if isinstance(tlw, QMainWindow):
                    parent = tlw
                    break

        # bail out if we're not running on Qt's application thread
        if not onAppThread():
            sys.stderr.write("excMessageBox: not on application thread; bailing out\n")
            return

        if not title:
            title = _("Unhandled exception")
        if not message:
            message = _("An exception was raised.")

        if showExcSummary:
            try:
                from gitfourchette import trtables
                summary = f"{trtables.exceptionName(exc)}: {exc}"
            except:  # noqa: E722
                summaryLines = traceback.format_exception_only(exc.__class__, exc)
                summary = ''.join(summaryLines).strip()

            if len(summary) > 500:
                summary = summary[:500] + "... " + _("(MESSAGE TRUNCATED)")
            message += "<br><br>" + html.escape(summary)
            if isCritical:
                message += "<p><small>" + _("If you want to file a bug report, please click “Show Details” "
                                            "and copy the <b>entire</b> message.")

            tracebackLines = traceback.format_exception(exc.__class__, exc, exc.__traceback__)
            details = ''.join(shortenTracebackPath(line) for line in tracebackLines).strip()

        qmb = asyncMessageBox(parent, icon, title, message)

        if showExcSummary:
            qmb.setDetailedText(details)
            detailsEdit: QTextEdit = qmb.findChild(QTextEdit)
            if detailsEdit:
                font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
                font.setPointSize(min(font.pointSize(), 8))
                detailsEdit.setFont(font)
                detailsEdit.setMinimumWidth(600)
                detailsEdit.setFixedHeight(300)

        # Keep user from triggering more exceptions by clicking on stuff in the background
        qmb.setWindowModality(Qt.WindowModality.ApplicationModal)

        if APP_DEBUG and isCritical:
            quitButton = qmb.addButton(QMessageBox.StandardButton.Reset)  # Reset button is leftmost in KDE
            quitButton.setText(_("Quit application"))

        dismissButton = qmb.addButton(QMessageBox.StandardButton.Ok)
        qmb.setDefaultButton(dismissButton)
        qmb.setEscapeButton(dismissButton)

        # Show next error message in queue when finished
        qmb.finished.connect(_popExcMessageBoxQueue)

        # Add the qmb to the queue of error messages
        _excMessageBoxQueue.append(qmb)

        # Only show now if no other excMessageBox is currently being shown
        if len(_excMessageBoxQueue) <= 1:
            _showExcMessageBox(qmb)

    except BaseException as excMessageBoxError:  # pragma: no cover
        sys.stderr.write("*********************************************\n")
        sys.stderr.write("excMessageBox failed!!!\n")
        sys.stderr.write("*********************************************\n")
        traceback.print_exception(excMessageBoxError)

    if APP_TESTMODE and abortUnitTest:
        raise exc


def _popExcMessageBoxQueue(result=QMessageBox.StandardButton.Ok):
    if result == QMessageBox.StandardButton.Reset:
        logger.warning("Application aborted from message box.")
        QApplication.exit(1)
        return

    if result != QMessageBox.StandardButton.Ok:
        for qmb in _excMessageBoxQueue:
            qmb.deleteLater()
        _excMessageBoxQueue.clear()
        return

    _excMessageBoxQueue.pop(0)
    if not _excMessageBoxQueue:
        # No more messages to show
        return

    qmb = _excMessageBoxQueue[0]

    numRemaining = len(_excMessageBoxQueue) - 1
    if numRemaining >= 1:
        dismissAllButton = qmb.addButton(QMessageBox.StandardButton.NoToAll)
        dismissAllButton.setText(_n("Skip {n} more error", "Skip {n} more errors", numRemaining))

    _showExcMessageBox(qmb)


def _showExcMessageBox(qmb):
    qmb.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)  # don't leak dialog

    if qmb.parent():
        qmb.show()
    else:
        # Without a parent, show() won't work. So, use exec() as the very last resort.
        # (Calling exec() may crash on macOS if another modal dialog is active.)
        qmb.exec()


def asyncMessageBox(
        parent: QWidget | None,
        icon: MessageBoxIconName,
        title: str,
        text: str,
        buttons=QMessageBox.StandardButton.NoButton,
        macShowTitle=True,
        deleteOnClose=True,
) -> QMessageBox:

    assert onAppThread()
    assert parent is None or isinstance(parent, QWidget)

    title = title.split("\u009C", 1)[0]
    text = text.split("\u009C", 1)[0]

    loggedMessage = f"[{title}] {stripHtml(text)}"
    if icon in ['information', 'question']:
        logger.debug(loggedMessage)
    else:
        logger.warning(loggedMessage)

    icons = {
        'warning': QMessageBox.Icon.Warning,
        'information': QMessageBox.Icon.Information,
        'question': QMessageBox.Icon.Question,
        'critical': QMessageBox.Icon.Critical,
    }

    # macOS doesn't have a titlebar for message boxes, so put the title in the text
    if MACOS and macShowTitle:  # pragma: no cover
        text = f"<p><b>{title}</b></p>{text}"

    qmb = QMessageBox(
        icons.get(icon, QMessageBox.Icon.NoIcon),
        title,
        text,
        buttons,
        parent=parent
    )

    if parent:
        qmb.setWindowModality(Qt.WindowModality.WindowModal)

    if deleteOnClose:
        qmb.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

    # macOS hacks to remove "modern" styling
    if MACOS:
        MakeNonNativeDialog(qmb)

    return qmb


def showWarning(parent: QWidget, title: str, text: str) -> QMessageBox:
    """
    Shows a warning message box asynchronously.
    """
    qmb = asyncMessageBox(parent, 'warning', title, text)
    qmb.show()
    return qmb


def showInformation(parent: QWidget, title: str, text: str) -> QMessageBox:
    """
    Shows an information message box asynchronously.
    """
    qmb = asyncMessageBox(parent, 'information', title, text)
    qmb.show()
    return qmb


def askConfirmation(
        parent: QWidget,
        title: str,
        text: str,
        callback: Callable | None = None,
        buttons=QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel,
        okButtonText: str = "",
        okButtonIcon: QIcon | None = None,
        show=True,
        messageBoxIcon: MessageBoxIconName = "question",
) -> QMessageBox:
    """
    Shows a confirmation message box asynchronously.

    If you override `buttons`, be careful with your choice of StandardButton values;
    some of them won't emit the `accepted` signal which is connected to the callback.
    """

    qmb = asyncMessageBox(parent, messageBoxIcon, title, text, buttons)

    okButton = qmb.button(QMessageBox.StandardButton.Ok)
    if okButton:
        if okButtonText:
            okButton.setText(okButtonText)
        if okButtonIcon:
            okButton.setIcon(okButtonIcon)

    if callback:
        qmb.accepted.connect(callback)

    if show:
        qmb.show()

    return qmb


def addULToMessageBox(qmb: QMessageBox, items: list[str], limit=10):
    ul = ulify(items, limit=limit, moreText=_("…and {0} more. Click “Show Details” to view all."))

    qmb.setText(qmb.text() + ul)

    if len(items) > limit:
        overflow = ulify(items, -1)
        qmb.setDetailedText(overflow)
        qte: QTextEdit = qmb.findChild(QTextEdit)
        if qte is not None:
            qte.setHtml(overflow)


class NonCriticalOperation:
    def __init__(self, operation: str):
        self.operation = operation

    def __enter__(self):
        pass

    def __exit__(self, excType, excValue, excTraceback):
        if excValue:
            excMessageBox(excValue, message=_("Operation failed: {0}.", html.escape(self.operation)))
            return True  # don't propagate



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.ui_commitinfodialog import Ui_CommitInfoDialog
from gitfourchette.localization import _
from gitfourchette.qt import *
from gitfourchette.toolbox.iconbank import stockIcon
from gitfourchette.toolbox.qtutils import packDialog


class CommitInfoDialog(QDialog):
    """
    Rich summary (HTML) plus optional read-only body, in a resizable dialog.
    Used instead of QMessageBox.setDetailedText because QMessageBox forces a
    fixed size whenever its layout updates.
    """

    _PreferredWidth = 512

    @property
    def summaryLabel(self):
        return self.ui.summaryLabel

    def __init__(
            self,
            parent: QWidget | None,
            title: str,
            summaryHtml: str,
            body: str = "",
    ):
        super().__init__(parent)

        self.ui = Ui_CommitInfoDialog()
        self.ui.setupUi(self)

        self.setWindowTitle(title)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setWindowModality(Qt.WindowModality.WindowModal)

        # Start without details box (we need to capture the packed size first)
        self.ui.detailsEdit.setVisible(False)

        pm = stockIcon("SP_MessageBoxInformation").pixmap(56, 56)
        self.ui.iconLabel.setPixmap(pm)
        self.ui.summaryLabel.setText(summaryHtml)
        self.ui.detailsEdit.setPlainText(body)

        self._detailsToggleButton = self.ui.buttonBox.addButton("...", QDialogButtonBox.ButtonRole.ActionRole)
        self._detailsToggleButton.clicked.connect(self._toggleDetailsPane)
        self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setFocus()

        # Pack the layout, then save the packed size
        self.setMinimumWidth(self._PreferredWidth)
        packDialog(self)
        self.packedSize = self.size()

        # Show details if any
        if body:
            self._toggleDetailsPane()  # pane currently hidden - show it
        else:
            self._detailsToggleButton.setVisible(False)

        self._applyResizePolicy()

    def isShowingDetails(self) -> bool:
        # isVisibleTo, not isVisible, so this is valid even before we're shown
        return self.ui.detailsEdit.isVisibleTo(self)

    def _toggleDetailsPane(self):
        show = not self.isShowingDetails()
        self.ui.detailsEdit.setVisible(show)
        self._detailsToggleButton.setText(_("Hide Full &Message") if show else _("Show Full &Message"))
        self._applyResizePolicy()

    def _applyResizePolicy(self):
        if self.isShowingDetails():
            self.setSizeGripEnabled(True)
            self.setMinimumSize(self._PreferredWidth, 0)
            self.setMaximumSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX)
            self.adjustSize()
        else:
            self.setSizeGripEnabled(False)
            self.setFixedSize(self.packedSize)



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Manage proprietary settings in a repository's .git/config and .git/gitfourchette.json.
"""

from dataclasses import dataclass, field

from gitfourchette.appconsts import *
from gitfourchette.forms.signatureform import SignatureOverride
from gitfourchette.porcelain import *
from gitfourchette.settings import RefSort
from gitfourchette.prefsfile import PrefsFile


@dataclass
class RepoPrefs(PrefsFile):
    _filename = f"{APP_SYSTEM_NAME}.json"
    _allowMakeDirs = False
    _parentDir = ""

    _GitConfigShadowUpstreamKey = f"{APP_SYSTEM_NAME}-shadow-remote"

    _repo: Repo
    draftCommitMessage: str = ""
    draftCommitSignature: Signature | None = None
    draftCommitSignatureOverride: SignatureOverride = SignatureOverride.Nothing
    draftAmendMessage: str = ""
    hidePatterns: set = field(default_factory=set)
    showPatterns: set = field(default_factory=set)
    collapseCache: set = field(default_factory=set)
    sortBranches: RefSort = RefSort.UseGlobalPref
    sortRemoteBranches: RefSort = RefSort.UseGlobalPref
    sortTags: RefSort = RefSort.UseGlobalPref
    refSortClearTimestamp: int = 0
    customKeyFile: str = ""

    @classmethod
    def initForRepo(cls, repo: Repo):
        repoPrefs = cls(repo)
        repoPrefs._parentDir = repo.path
        repoPrefs.load()
        return repoPrefs

    def getParentDir(self):
        return self._parentDir

    def hasDraftCommit(self):
        return self.draftCommitMessage or self.draftCommitSignatureOverride != SignatureOverride.Nothing

    def clearDraftCommit(self):
        self.draftCommitMessage = ""
        self.draftCommitSignature = None
        self.draftCommitSignatureOverride = SignatureOverride.Nothing
        self.setDirty()

    def clearDraftAmend(self):
        self.draftAmendMessage = ""
        self.setDirty()

    def clearRefSort(self):
        self.sortBranches = RefSort.UseGlobalPref
        self.sortRemoteBranches = RefSort.UseGlobalPref
        self.sortTags = RefSort.UseGlobalPref
        self.refSortClearTimestamp = 0
        self.setDirty()

    def getShadowUpstream(self, localBranchName: str):
        return self._repo.get_config_value(("branch", localBranchName, self._GitConfigShadowUpstreamKey))

    def setShadowUpstream(self, localBranchName: str, upstreamName: str):
        self._repo.set_config_value(("branch", localBranchName, self._GitConfigShadowUpstreamKey), upstreamName)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.blameview.blamemodel import BlameModel
from gitfourchette.blameview.blamescrubberdelegate import BlameScrubberDelegate
from gitfourchette.blameview.blamescrubbermodel import BlameScrubberModel
from gitfourchette.qt import *
from gitfourchette.toolbox import enforceComboBoxMaxVisibleItems


class BlameScrubber(QComboBox):
    def __init__(self, blameModel: BlameModel, parent: QWidget):
        super().__init__(parent)

        self.scrubberDelegate = BlameScrubberDelegate(blameModel, parent=self)
        self.scrubberModel = BlameScrubberModel(blameModel, parent=self)

        self.setMinimumWidth(128)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)

        # For performance, prevent Qt from looking at every item in the model during initialization
        self.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)

        view = self.view()
        assert isinstance(view, QListView)
        view.setUniformItemSizes(True)

        self.setItemDelegate(self.scrubberDelegate)
        self.setModel(self.scrubberModel)

        enforceComboBoxMaxVisibleItems(self, 25)

    def paintEvent(self, e):
        painter = QStylePainter(self)

        controlOption = QStyleOptionComboBox()
        self.initStyleOption(controlOption)
        painter.drawComplexControl(QStyle.ComplexControl.CC_ComboBox, controlOption)

        rect = painter.style().subElementRect(QStyle.SubElement.SE_ComboBoxFocusRect, controlOption, self)

        itemOption = QStyleOptionViewItem()
        itemOption.initFrom(self)
        itemOption.widget = self
        itemOption.rect = rect
        modelIndex = self.model().index(self.currentIndex(), 0)

        delegate = self.scrubberDelegate
        delegate.singleItem = True
        delegate.paint(painter, itemOption, modelIndex, fillBackground=False)
        delegate.singleItem = False



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from gitfourchette import trtables
from gitfourchette.graph.graph import CommitTraits
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID, RepoModel, GpgStatus
from gitfourchette.toolbox import *


@dataclass
class CommitToolTipZone:
    left: int
    right: int
    kind: Literal['ref', 'author', 'message', 'pathspec']
    data: str = ""


class SpecialRow(enum.IntEnum):
    Invalid = 0

    # Dedicated rows in the commit log
    UncommittedChanges = enum.auto()
    Commit = enum.auto()
    TruncatedHistory = enum.auto()
    EndOfShallowHistory = enum.auto()

    # The items below are not dedicated rows in the commit log per se.
    # They are special kinds of SpecialDiffError displayed when an invalid
    # selection is made in GraphView.
    TooManyRowsSelected = enum.auto()
    CannotCompareRows = enum.auto()

    @classmethod
    def fromString(cls, s: str) -> SpecialRow:
        i = int(s)
        return cls(i)


class CommitLogModel(QAbstractListModel):
    ToolTipCacheSize = 150
    """ Number of rows to keep track of for ToolTipZones """

    class Role:
        Commit          = Qt.ItemDataRole.UserRole + 0
        Oid             = Qt.ItemDataRole.UserRole + 1
        ToolTipZones    = Qt.ItemDataRole.UserRole + 2
        AuthorColumnX   = Qt.ItemDataRole.UserRole + 3
        SpecialRow      = Qt.ItemDataRole.UserRole + 4
        BlameRevision   = Qt.ItemDataRole.UserRole + 5  # for BlameScrubber
        ComparisonSide  = Qt.ItemDataRole.UserRole + 6  # A/B commit diffs

    repoModel: RepoModel
    _extraRow: SpecialRow
    _authorColumnX: int
    _toolTipZones: dict[int, list[CommitToolTipZone]]

    def __init__(self, repoModel: RepoModel, parent: QWidget):
        super().__init__(parent)

        self.repoModel = repoModel
        self._authorColumnX = -1
        self._toolTipZones = {}
        self.commitDiffAB: tuple[Oid, Oid] | None = None

        if repoModel.truncatedHistory:
            self._extraRow = SpecialRow.TruncatedHistory
        elif repoModel.repo.is_shallow:
            self._extraRow = SpecialRow.EndOfShallowHistory
        else:
            self._extraRow = SpecialRow.Invalid

    def resetCommitSequence(self, nRemovedRows: int = -1, nAddedRows: int = 0):
        if nRemovedRows < 0:
            # Replace log wholesale
            self.beginResetModel()
            self.endResetModel()
            return

        parent = QModelIndex_default  # it's not a tree model so there's no parent

        # DON'T interleave beginRemoveRows/beginInsertRows!
        # It'll crash with QSortFilterProxyModel!
        if nRemovedRows != 0:
            self.beginRemoveRows(parent, 0, nRemovedRows)
            self.endRemoveRows()

        if nAddedRows != 0:
            self.beginInsertRows(parent, 0, nAddedRows)
            self.endInsertRows()

    def rowCount(self, *args, **kwargs) -> int:
        n = len(self.repoModel.commitSequence)
        if self._extraRow != SpecialRow.Invalid:
            n += 1
        return n

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        row = index.row()

        if role == Qt.ItemDataRole.DisplayRole:
            return None

        elif role == CommitLogModel.Role.Commit:
            try:
                return self.repoModel.commitSequence[row]
            except IndexError:
                return None

        elif role == CommitLogModel.Role.Oid:
            try:
                commit = self.repoModel.commitSequence[row]
                return commit.id
            except (IndexError, AttributeError):
                return None

        elif role == CommitLogModel.Role.SpecialRow:
            if row == 0:
                return SpecialRow.UncommittedChanges
            elif row < len(self.repoModel.commitSequence):
                return SpecialRow.Commit
            else:
                return self._extraRow

        elif role == CommitLogModel.Role.ComparisonSide:
            if not self.commitDiffAB:
                return ""
            try:
                commit = self.repoModel.commitSequence[row]
                i = self.commitDiffAB.index(commit.id)
                return "AB"[i]
            except (IndexError, AttributeError, ValueError):
                return ""

        elif role == Qt.ItemDataRole.ToolTipRole:
            tip = ""

            try:
                commit = self.repoModel.commitSequence[row]
                zones = self._toolTipZones[row]
            except (IndexError, KeyError):
                return tip
            if commit is None or commit.id == UC_FAKEID:
                return tip

            parentWidget = self.parent()
            assert isinstance(parentWidget, QWidget)
            x = parentWidget.mapFromGlobal(QCursor.pos()).x()

            for zone in reversed(zones):
                if not (zone.left <= x <= zone.right):
                    continue
                if zone.kind == "ref":
                    tip = zone.data
                elif zone.kind == "message":
                    tip = commitMessageTooltip(commit)
                elif zone.kind == "author":
                    tip = commitAuthorTooltip(commit, *self.repoModel.getCachedGpgStatus(commit))
                elif zone.kind == "pathspec":
                    tip = _("This commit touches a path that matches your search")
                break

            if self._authorColumnX <= 0:  # author hidden in narrow window
                tip += commitAuthorTooltip(commit, *self.repoModel.getCachedGpgStatus(commit))

            return tip

        return None

    def setData(self, index, value, role=None):
        if role == CommitLogModel.Role.AuthorColumnX:
            self._authorColumnX = value
            return True

        elif role == CommitLogModel.Role.ToolTipZones:
            row = index.row()

            self._toolTipZones.pop(row, None)

            if value:
                # Bump row to end of keys (dicts keep key insertion order since Python 3.7)
                self._toolTipZones[row] = value

                # Nuke old entries if the dict grew beyond the threshold
                trimCacheDict(self._toolTipZones, CommitLogModel.ToolTipCacheSize)

            return True

        return False


def commitAuthorTooltip(commit: CommitTraits, gpgStatus: GpgStatus, gpgKeyInfo: str) -> str:
    def formatTime(sig: Signature):
        return escape(signatureDateFormat(sig))

    def formatPerson(sig: Signature):
        return f"<b>{escape(sig.name)}</b> &lt;{escape(sig.email)}&gt;"

    author = commit.author
    committer = commit.committer

    markup = "<p style='white-space: pre'>"
    markup += formatPerson(author)

    if author == committer:
        markup += f"<small><br>{formatTime(author)}"
    elif author.name == committer.name and author.email == committer.email:
        suffixA = _p("CommitTooltip", "(authored)")
        suffixC = _p("CommitTooltip", "(committed)")
        markup += (f"<small><br>{formatTime(author)} {suffixA}"
                   f"<br>{formatTime(committer)} *{suffixC}")
    else:
        committedBy = _p("CommitTooltip", "Committed by {0}", formatPerson(committer))
        markup += (f"<small><br>{formatTime(author)}"
                   f"<br><br>* {committedBy}"
                   f"<br><small>{formatTime(committer)}")
    markup += "</p>"

    if gpgStatus == GpgStatus.Unsigned:
        pass
    elif gpgKeyInfo:
        markup += f"<p>{gpgStatus.iconHtml()} {trtables.enum(gpgStatus)}<br><small>{escape(gpgKeyInfo)}</small></p>"
    else:
        markup += f"<p>{gpgStatus.iconHtml()} {trtables.enum(gpgStatus)}</p>"

    return markup


def commitMessageTooltip(commit: CommitTraits) -> str:
    message = commit.message.rstrip()
    maxLength = max(len(line) for line in message.splitlines())
    message = escape(message)

    if maxLength <= 80:
        # Keep Qt from wrapping tooltip text when the message is made up of short lines
        return "<p style='white-space: pre'>" + message
    else:
        return "<p>" + message.replace('\n', '<br>')


def trimCacheDict(d: dict, trimToSize: int):
    maxCapacity = trimToSize * 2
    size = len(d)
    if size <= maxCapacity:
        return
    numOldKeys = size - trimToSize
    oldKeys = list(d)[:numOldKeys]
    for k in oldKeys:
        del d[k]



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *


class QComboBoxWithPreview(QComboBox):
    dataPicked = Signal(object)

    class Role:
        Data = Qt.ItemDataRole.UserRole + 0
        Preview = Qt.ItemDataRole.UserRole + 1

    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.numPresets = 0
        self.captionWidth = 0
        self.previewWidth = 0
        delegate = QComboBoxWithPreviewDelegate(self)
        self.setItemDelegate(delegate)
        self.activated.connect(self.onActivated)

    def addItemWithPreview(self, caption: str, data: object, preview: str):
        i = self.count()
        self.addItem(caption)
        self.setItemData(i, data, QComboBoxWithPreview.Role.Data)
        self.setItemData(i, preview, QComboBoxWithPreview.Role.Preview)

        fontMetrics = self.fontMetrics()
        self.captionWidth = max(self.captionWidth, fontMetrics.horizontalAdvance(caption) + 20)
        self.previewWidth = max(self.previewWidth, fontMetrics.horizontalAdvance(preview) + 15)

        self.numPresets += 1

    def showPopup(self):
        # TODO: Where is QListView padding defined?
        self.view().setMinimumWidth(3 + self.captionWidth + self.previewWidth + 3)
        super().showPopup()

    def onActivated(self, index: int):
        # The signal may be sent for an index beyond the number of presets
        # when the user hits enter with a custom item.
        if index < 0 or index >= self.numPresets:
            return

        data = self.itemData(index, QComboBoxWithPreview.Role.Data)
        self.dataPicked.emit(data)

        if self.isEditable():
            self.setEditText(str(data))


class QComboBoxWithPreviewDelegate(QStyledItemDelegate):
    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex):
        super().paint(painter, option, index)
        painter.save()

        pw = self.parent()
        assert isinstance(pw, QComboBoxWithPreview)

        rect = QRect(option.rect)
        rect.setLeft(rect.left() + 3 + pw.captionWidth)

        isSelected = bool(option.state & QStyle.StateFlag.State_Selected)
        colorRole = QPalette.ColorRole.PlaceholderText if not isSelected else QPalette.ColorRole.HighlightedText

        font: QFont = painter.font()
        font.setItalic(True)

        preview = index.data(QComboBoxWithPreview.Role.Preview)

        painter.setFont(font)
        painter.setPen(option.palette.color(QPalette.ColorGroup.Normal, colorRole))
        painter.drawText(rect, Qt.AlignmentFlag.AlignVCenter, str(preview))
        painter.restore()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import json
import sys
from pathlib import Path
from textwrap import dedent

import pygit2
import pygments

from gitfourchette.forms.prefsdialog import availableLocaleCodes, localeCodeToLanguageName
from gitfourchette.forms.ui_aboutdialog import Ui_AboutDialog
from gitfourchette.gitdriver import GitDriver
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *
from gitfourchette import settings

WEBSITE_URL = "https://gitfourchette.org"
DONATE_URL = "https://ko-fi.com/jorio"
TRANSLATE_URL = "https://gitfourchette.org/localization"


def getPygit2FeatureStrings():
    return [f.name.lower() for f in pygit2.enums.Feature if f & pygit2.features]


def simpleLink(url):
    return f"<a href='{url}'>{url}</a>"


class AboutDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        self.ui = Ui_AboutDialog()
        self.ui.setupUi(self)

        self.urlToolTip = UrlToolTip(self)
        self.urlToolTip.install()

        appVersion = QApplication.applicationVersion()
        appName = qAppName()

        self.setWindowTitle(self.windowTitle().format(appName))

        buildInfoItems = [
            "Flatpak" if FLATPAK else "",
            "AppImage" if "APPIMAGE" in INITIAL_ENVIRONMENT else "",
            APP_FREEZE_DATE,
            APP_FREEZE_COMMIT[:7]
        ]
        buildInfoItems = [s for s in buildInfoItems if s]
        buildInfo = f"({', '.join(buildInfoItems)})" if buildInfoItems else ""

        tagline = _("The comfortable Git UI for Linux.")

        # ---------------------------------------------------------------------
        # Header

        pixmap = QPixmap("assets:icons/gitfourchette")
        pixmap.setDevicePixelRatio(4)
        self.ui.iconLabel.setPixmap(pixmap)

        self.ui.header.setText(dedent(f"""\
            <span style="font-size: x-large"><b>{appName}</b></span>
            <br>{tagline}
            <br>{simpleLink(WEBSITE_URL)}"""))

        versionText = _("Version {0}", appVersion)
        self.ui.versionLabel.setText(dedent(f"""\
            <span style='color:{mutedTextColorHex(self)}'><b>{versionText}</b> {buildInfo}
            <br>Copyright © 2026 Iliyas Jorio"""))

        # ---------------------------------------------------------------------
        # About page

        self.ui.mugshot.setText("")
        self.ui.mugshot.setPixmap(QPixmap("assets:icons/mug"))

        # ---------------------------------------------------------------------
        # Components page

        try:
            from gitfourchette.mount.treemount import fuse
            fuseInfo = _("with {0}", f"FUSE {fuse.fuse_version_major}.{fuse.fuse_version_minor}")
        except (ImportError, OSError, AttributeError):
            fuseInfo = _("(not available)")

        self.ui.aboutBlurb.setText(dedent(f"""\
            <style>
                a {{ font-weight: bold; }}
                hr {{ background: gray; }}
                table {{ font-size: 8pt; margin-top: 0em; }}
                th {{ text-align: left; padding-right: 1em; color: gray; }}
            </style>
            <p>{linkify(_("If {app} helps you get work done, please consider [making a small donation].", app=appName), DONATE_URL)}</p>
            <p>{_("Thank you for your support!")}</p>
            <hr>
            <table cellspacing=0 cellpadding=0>
            <tr><th colspan=2>{appName} {appVersion}</th></tr>
            <tr><th>Git</th><td>{GitDriver.gitVersion() or _("(unknown version)")} ({settings.prefs.gitPath})</td></tr>
            <tr><th>LFS</th><td>{GitDriver.lfsVersion() or _("(not available)")}</td></tr>
            <tr><th>pygit2</th><td>{pygit2.__version__}</td></tr>
            <tr><th>libgit2</th><td>{pygit2.LIBGIT2_VERSION} ({', '.join(getPygit2FeatureStrings())})</td></tr>
            <tr><th>{QT_BINDING}</th><td>{QT_BINDING_VERSION}</td></tr>
            <tr><th>Qt</th><td>{qVersion()}</td></tr>
            <tr><th>mfusepy</th><td>{fuseInfo}</td></tr>
            <tr><th>Pygments</th><td>{pygments.__version__}</td></tr>
            <tr><th>Python</th><td>{'.'.join(str(i) for i in sys.version_info)}</td></tr>
            </table>
        """))

        # ---------------------------------------------------------------------
        # Acknowledgments page

        contributorCreditsPath = Path(QFile("assets:/contributors.txt").fileName())
        contributorCredits = contributorCreditsPath.read_text("utf-8")
        contributorCredits = contributorCredits.removeprefix("Iliyas Jorio\n")
        contributorCredits = f"<center style='white-space: pre'>{contributorCredits}</center>"

        translatorCreditsPath = Path(QFile("assets:lang/credits.json").fileName())
        translatorCredits = json.loads(translatorCreditsPath.read_text("utf-8"))
        translatorFilter = availableLocaleCodes()
        translatorMarkup = "<center><table>" + "".join(
            f"<tr><th>{localeCodeToLanguageName(lang)}:</th><td>{'<br>'.join(people)}</td></tr>\n"
            for lang, people in translatorCredits.items()
            if lang in translatorFilter
        ) + "</table></center>"

        ackText = [
            _("Additional contributions by:") + contributorCredits,

            linkify(_("Translations welcome!"), TRANSLATE_URL)
            + " "
            + _("Brought to your native language by:")
            + translatorMarkup,

            _("Special thanks to Marc-Alexandre Espiaut for beta testing."),
        ]

        ackMarkup = "<style>th { text-align: right; padding-right: 8px; color: gray; }</style>"
        ackMarkup += paragraphs(*ackText)

        self.ui.ackBlurb.setText(ackMarkup)

        # ---------------------------------------------------------------------
        # License page

        self.ui.licenseBlurb.setText(dedent(f"""\
            <p>{appName} is free software: you can redistribute it and/or
            modify it under the terms of the GNU General Public License
            version 3 as published by the Free Software Foundation.</p>
            <p>{appName} is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. For more
            details, read the full terms of the
            <a href='https://www.gnu.org/licenses/gpl-3.0.txt'>GNU General
            Public License, version 3</a>.</p>"""))

    @staticmethod
    def popUp(parent: QWidget):
        dialog = AboutDialog(parent)
        dialog.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        dialog.show()
        return dialog



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class ProtocolButton(QToolButton):
    protocolChanged = Signal(str)

    def __init__(self, parent):
        super().__init__(parent)
        self.setFixedWidth(self.sizeHint().width() + self.fontMetrics().horizontalAdvance("--https--"))
        self.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.setToolTip(_("Change URL protocol"))

    def connectTo(self, lineEdit: QLineEdit):
        def onUrlProtocolChanged(newUrl: str):
            # This pushes the new text to the QLineEdit's undo stack
            # (whereas setText clears the undo stack).
            lineEdit.selectAll()
            lineEdit.insert(newUrl)

        lineEdit.textChanged.connect(self.onUrlChanged)
        self.protocolChanged.connect(onUrlProtocolChanged)

        # Prime state
        self.onUrlChanged(lineEdit.text())

    def onUrlChanged(self, url: str):
        """ Detect protocol when URL changes """

        url = url.strip()
        protocol = remoteUrlProtocol(url)

        if not protocol:  # unknown protocol, hide protocol button
            self.hide()
            return

        # Build alternate URL
        host, path = splitRemoteUrl(url)
        if protocol == "ssh":
            newUrl = f"https://{host}/{path}"
            newUrl = newUrl.removesuffix(".git")
        else:
            host = host.split(":", 1)[0]  # remove port, if any
            newUrl = f"git@{host}:{path}"

        self.show()
        self.setText(protocol)

        menu = ActionDef.makeQMenu(self, [ActionDef(newUrl, lambda: self.protocolChanged.emit(newUrl))])
        self.setMenu(menu)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import os
import shlex
from contextlib import suppress
from typing import ClassVar

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette import tasks
from gitfourchette.diffarea import DiffArea
from gitfourchette.exttools.toolprocess import ToolProcess
from gitfourchette.exttools.usercommand import UserCommand
from gitfourchette.forms.banner import Banner
from gitfourchette.forms.processdialog import ProcessDialog
from gitfourchette.forms.repostub import RepoStub
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.graphview.graphview import GraphView
from gitfourchette.localization import *
from gitfourchette.nav import NavHistory, NavLocator, NavContext
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel, UC_FAKEID
from gitfourchette.sidebar.sidebar import Sidebar
from gitfourchette.syntax import LexJobCache
from gitfourchette.tasks import RepoTaskRunner, TaskEffects, TaskBook
from gitfourchette.tasks.misctasks import VerifyGpgQueue
from gitfourchette.tasks.nettasks import AutoFetchRemotes
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class RepoWidget(QWidget):
    sharedSplitterSizes: ClassVar[dict[str, list[int]]] = {}
    "Shared reference among all RepoWidgets"

    nameChange = Signal()
    openRepo = Signal(str, NavLocator)
    openPrefs = Signal(str)
    locatorChanged = Signal(NavLocator)
    historyChanged = Signal()
    requestAttention = Signal()
    becameVisible = Signal()
    mustReplaceWithStub = Signal(RepoStub)
    aboutToDelete = Signal()

    busyMessage = Signal(str)
    statusMessage = Signal(str)
    clearStatus = Signal()

    repoModel: RepoModel
    taskRunner: RepoTaskRunner

    navLocator: NavLocator
    navHistory: NavHistory

    splittersToSave: list[QSplitter]
    centralSplitSizesBackup: list[int]

    @property
    def repo(self) -> Repo:
        return self.repoModel.repo

    @property
    def workdir(self):
        return os.path.normpath(self.repoModel.repo.workdir)

    @property
    def superproject(self):
        return self.repoModel.superproject

    def __init__(self, repoModel: RepoModel, taskRunner: RepoTaskRunner, parent: QWidget):
        super().__init__(parent)
        self.setObjectName(f"{type(self).__name__}({repoModel.shortName})")

        # The stylesheet must be refreshed so that subsequent tweakFont calls can take effect.
        reevaluateStyleSheet(self)

        # Use RepoTaskRunner to schedule git operations to run on a separate thread.
        self.taskRunner = taskRunner
        self.taskRunner.setParent(self)
        self.taskRunner.ready.connect(self.onTaskRunnerReady)
        self.taskRunner.progress.connect(self.onRepoTaskProgress)
        self.taskRunner.repoGone.connect(self.onRepoGone)
        self.taskRunner.requestAttention.connect(self.requestAttention)

        # Report progress in long-running background processes
        self.processDialog = ProcessDialog(self)
        self.taskRunner.processStarted.connect(self.processDialog.connectProcess)

        self.repoModel = repoModel
        self.lastAutoFetchTime = QDateTime.currentSecsSinceEpoch()

        self.busyCursorDelayer = QTimer(self)
        self.busyCursorDelayer.setSingleShot(True)
        self.busyCursorDelayer.setInterval(100)
        self.busyCursorDelayer.timeout.connect(self.onBusyCursorDelayerTimeout)

        self.navLocator = NavLocator()
        self.navHistory = NavHistory()

        self.centralSplitSizesBackup = []

        # ----------------------------------
        # Splitters

        sideSplitter = QSplitter(Qt.Orientation.Horizontal, self)
        sideSplitter.setObjectName("Split_Side")
        self.sideSplitter = sideSplitter

        centralSplitter = QSplitter(Qt.Orientation.Vertical, self)
        centralSplitter.setObjectName("Split_Central")
        self.centralSplitter = centralSplitter

        dummyLayout = QVBoxLayout()
        dummyLayout.setSpacing(0)
        dummyLayout.setContentsMargins(QMargins())
        dummyLayout.addWidget(sideSplitter)
        self.setLayout(dummyLayout)

        # ----------------------------------
        # Build widgets

        sidebarContainer = self._makeSidebarContainer()
        graphContainer = self._makeGraphContainer()

        self.diffArea = DiffArea(self.repoModel, self)
        # Bridges for legacy code
        self.dirtyFiles = self.diffArea.dirtyFiles
        self.stagedFiles = self.diffArea.stagedFiles
        self.committedFiles = self.diffArea.committedFiles
        self.diffView = self.diffArea.diffView
        self.specialDiffView = self.diffArea.specialDiffView
        self.conflictView = self.diffArea.conflictView
        self.diffBanner = self.diffArea.diffBanner

        # ----------------------------------
        # Add widgets in splitters

        sideSplitter.addWidget(sidebarContainer)
        sideSplitter.addWidget(centralSplitter)
        sideSplitter.setSizes([220, 500])
        sideSplitter.setStretchFactor(0, 0)  # don't auto-stretch sidebar when resizing window
        sideSplitter.setStretchFactor(1, 1)
        sideSplitter.setChildrenCollapsible(False)

        centralSplitter.addWidget(graphContainer)
        centralSplitter.addWidget(self.diffArea)
        centralSplitter.setSizes([100, 150])
        centralSplitter.setCollapsible(0, True)  # Let DiffArea be maximized, thereby hiding the graph
        centralSplitter.setCollapsible(1, False)  # DiffArea can never be collapsed
        self.centralSplitSizesBackup = centralSplitter.sizes()
        self.diffArea.contextHeader.maximizeButton.clicked.connect(self.maximizeDiffArea)
        centralSplitter.splitterMoved.connect(self.syncDiffAreaMaximizeButton)

        splitters: list[QSplitter] = self.findChildren(QSplitter)
        assert all(s.objectName() for s in splitters), "all splitters must be named, or state saving won't work!"
        self.splittersToSave = splitters

        # ----------------------------------
        # Connect signals

        # save splitter state in splitterMoved signal
        for splitter in self.splittersToSave:
            splitter.splitterMoved.connect(lambda pos, index, s=splitter: self.saveSplitterState(s))

        for fileList in self.dirtyFiles, self.stagedFiles, self.committedFiles:
            # File list view selections are mutually exclusive.
            fileList.nothingClicked.connect(lambda: self.diffArea.clearDocument(NavLocator.inWorkdir()))
            fileList.statusMessage.connect(self.statusMessage)
            fileList.openSubRepo.connect(lambda path: self.openRepo.emit(self.repo.in_workdir(path), NavLocator()))

        self.graphView.linkActivated.connect(self.processInternalLink)
        self.graphView.statusMessage.connect(self.statusMessage)
        self.graphView.clDelegate.requestSignatureVerification.connect(self.repoModel.queueGpgVerification)
        self.graphView.clDelegate.requestSignatureVerification.connect(self.scheduleFlushGpgVerificationQueue)

        self.diffArea.conflictView.openPrefs.connect(self.openPrefs)
        self.diffArea.diffView.contextualHelp.connect(self.statusMessage)
        self.diffArea.specialDiffView.linkActivated.connect(self.processInternalLink)

        self.sidebar.statusMessage.connect(self.statusMessage)
        self.sidebar.toggleHideRefPattern.connect(self.toggleHideRefPattern)
        self.sidebar.openSubmoduleRepo.connect(self.openSubmoduleRepo)
        self.sidebar.openSubmoduleFolder.connect(self.openSubmoduleFolder)

        self.nameChange.connect(self.refreshWindowTitle)
        self.nameChange.connect(self.sidebar.sidebarModel.refreshRepoName)

        # ----------------------------------
        # Styling

        # Remove sidebar frame
        self.sidebar.setFrameStyle(QFrame.Shape.NoFrame)

        # Smaller fonts in diffArea buttons
        self.diffArea.applyCustomStyling()

        setTabOrder(
            self.sidebar.searchBar.lineEdit,
            self.sidebar,
            self.graphView.searchBar.lineEdit,
            self.graphView,
            self.diffArea.committedFiles.searchBar.lineEdit,
            self.diffArea.committedFiles,
            self.diffArea.dirtyFiles.searchBar.lineEdit,
            self.diffArea.dirtyFiles,
            self.diffArea.stagedFiles.searchBar.lineEdit,
            self.diffArea.stagedFiles,
            self.diffArea.diffView.searchBar.lineEdit,
            self.diffArea.diffView,
            self.diffArea.specialDiffView,
            self.diffArea.conflictView,
        )

        # ----------------------------------
        # Prime GraphView

        with QSignalBlockerContext(self.graphView):
            self.graphView.selectRowForLocator(NavLocator.inWorkdir())

        # ----------------------------------
        # Prime Sidebar

        with QSignalBlockerContext(self.sidebar):
            collapseCache = repoModel.prefs.collapseCache
            if collapseCache:
                self.sidebar.sidebarModel.collapseCache.update(collapseCache)
            self.sidebar.refresh(repoModel)

        # ----------------------------------
        self.restoreSplitterStates()
        self.refreshWindowTitle()
        self.refreshBanner()

        # Every second, check if we should auto-fetch.
        self.autoFetchTimer = QTimer(self)
        self.autoFetchTimer.timeout.connect(self.onAutoFetchTimerTimeout)
        self.autoFetchTimer.setInterval(1000)
        self.autoFetchTimer.start()

    def replaceWithStub(
            self,
            locator: NavLocator = NavLocator.Empty,
            maxCommits: int = -1,
            message: str = ""
    ) -> RepoStub:
        locator = locator or self.navLocator
        stub = RepoStub(parent=self.window(), workdir=self.workdir,
                        locator=locator, maxCommits=maxCommits)
        if message:
            stub.disableAutoLoad(message)
        self.mustReplaceWithStub.emit(stub)
        return stub

    def overridePendingLocator(self, locator: NavLocator):
        self.taskRunner.pendingEpilog.jumpTo = locator

    # -------------------------------------------------------------------------
    # Initial layout

    def _makeGraphContainer(self):
        graphView = GraphView(self.repoModel, self)

        container = QWidget()
        layout = QVBoxLayout(container)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(graphView.searchBar)
        layout.addWidget(graphView)

        self.graphView = graphView
        return container

    def _makeSidebarContainer(self):
        sidebar = Sidebar(self)

        banner = Banner(self, orientation=Qt.Orientation.Vertical)
        banner.setProperty("class", "merge")
        banner.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)

        container = QWidget()
        layout = QVBoxLayout(container)
        layout.setContentsMargins(QMargins())
        layout.setSpacing(0)
        layout.addWidget(sidebar.searchBar)
        layout.addWidget(sidebar)
        layout.addWidget(banner)

        self.sidebar = sidebar
        self.mergeBanner = banner

        return container

    # -------------------------------------------------------------------------
    # Splitter state

    def saveSplitterState(self, splitter: QSplitter):
        # QSplitter.saveState() saves a bunch of properties that we may want to
        # override in later versions, such as whether child widgets are
        # collapsible, the width of the splitter handle, etc. So, don't use
        # saveState(); instead, save the raw sizes for predictable results.
        name = splitter.objectName()
        sizes = splitter.sizes()[:]
        self.sharedSplitterSizes[name] = sizes

    def restoreSplitterStates(self):
        for splitter in self.splittersToSave:
            with suppress(KeyError):
                name = splitter.objectName()
                sizes = self.sharedSplitterSizes[name]
                splitter.setSizes(sizes)
        self.syncDiffAreaMaximizeButton()

    def isDiffAreaMaximized(self):
        sizes = self.centralSplitter.sizes()
        return sizes[0] == 0

    def maximizeDiffArea(self):
        if self.isDiffAreaMaximized():
            # Diff area was maximized - restore non-collapsed sizes
            newSizes = self.centralSplitSizesBackup
        else:
            # Maximize diff area - back up current sizes
            self.centralSplitSizesBackup = self.centralSplitter.sizes()
            newSizes = [0, 1]
        self.centralSplitter.setSizes(newSizes)
        self.saveSplitterState(self.centralSplitter)
        self.syncDiffAreaMaximizeButton()

    def syncDiffAreaMaximizeButton(self):
        isMaximized = self.isDiffAreaMaximized()
        self.diffArea.contextHeader.maximizeButton.setChecked(isMaximized)

    # -------------------------------------------------------------------------
    # Navigation

    def saveFilePositions(self):
        if self.diffView.isVisibleTo(self):
            newLocator = self.diffView.preciseLocator()
            if not newLocator.isSimilarEnoughTo(self.navLocator):
                warnings.warn(f"RepoWidget/DiffView locator mismatch: {self.navLocator} vs. {newLocator}")
        else:
            newLocator = self.navLocator.coarse()

        self.navHistory.push(newLocator)
        self.navLocator = newLocator

    def jump(self, locator: NavLocator, check=False):
        tasks.Jump.invoke(self, locator)
        if check:
            self.taskRunner.joinWorkerThread()
            assert self.navLocator.isSimilarEnoughTo(locator), f"failed to jump to: {locator}"

    def navigateBack(self):
        tasks.JumpBack.invoke(self)

    def navigateForward(self):
        tasks.JumpForward.invoke(self)

    # -------------------------------------------------------------------------

    def getTitle(self) -> str:
        return self.repoModel.shortName

    def closeEvent(self, event: QCloseEvent):
        """ Called when closing a repo tab """
        try:
            self.prepareForDeletion()
        except Exception as exc:  # pragma: no cover
            excMessageBox(exc, abortUnitTest=True)
        return super().closeEvent(event)

    def showEvent(self, event: QShowEvent):
        super().showEvent(event)
        self.becameVisible.emit()

    def prepareForDeletion(self):
        assert onAppThread()
        assert not hasattr(self, "_dead"), "RepoWidget already dead"
        self._dead = True

        # Kill any ongoing task then block UI thread until the task dies cleanly
        self.taskRunner.prepareForDeletion()

        self.aboutToDelete.emit()

        # Save sidebar collapse cache
        with NonCriticalOperation("Write repo prefs"):  # May raise OSError
            uiPrefs = self.repoModel.prefs
            collapseCache = self.sidebar.sidebarModel.collapseCacheLayers[0]
            if uiPrefs.collapseCache != collapseCache:
                uiPrefs.collapseCache = collapseCache.copy()
                uiPrefs.setDirty()
            if uiPrefs.isDirty():
                uiPrefs.write()

        # -----------------------------
        # GC help

        # Detangle cross-references to help out garbage collector
        self.diffView.gutter = None
        self.sidebar.repoWidget = None
        self.graphView.repoWidget = None
        for searchBar in self.findChildren(SearchBar):  # Help collect FileLists, GraphView, DiffView
            searchBar.buddy = None
        # Release any LexJobs that we own (it's not a big deal if we lose cached jobs for other tabs)
        LexJobCache.clear()

    def blameFile(self, path="", atCommit=NULL_OID):
        # Path not specified: pick one from the current locator
        if not path:
            loc = self.navLocator
            path = loc.path
            atCommit = self.navLocator.commit

            # If it's an uncommitted rename, start tracing the file's history
            # from its old name.
            if loc.context.isWorkdir():
                assert atCommit == NULL_OID
                delta = self.repoModel.findWorkdirDelta(loc.path)
                if delta is not None:
                    path = delta.old.path

        if not path:
            showInformation(self, tasks.OpenBlame.name(), _("Please select a file before performing this action."))
            return

        tasks.OpenBlame.invoke(self, path, atCommit)

    def openSubmoduleRepo(self, submoduleKey: str):
        path = self.repo.get_submodule_workdir(submoduleKey)
        self.openRepo.emit(path, NavLocator())

    def openSubmoduleFolder(self, submoduleKey: str):
        path = self.repo.get_submodule_workdir(submoduleKey)
        openFolder(path)

    def openRepoFolder(self):
        openFolder(self.workdir)

    def openSuperproject(self):
        superproject = self.superproject
        if superproject:
            self.openRepo.emit(superproject, NavLocator())
        else:
            showInformation(self, _("Open Superproject"), _("This repository does not have a superproject."))

    def copyRepoPath(self):
        text = self.workdir
        QApplication.clipboard().setText(text)
        self.statusMessage.emit(clipboardStatusMessage(text))

    def openGitignore(self):
        path = self.repo.in_workdir(".gitignore")
        self._openLocalConfigFile(path)

    def openLocalConfig(self):
        path = self.repo.in_gitdir("config")
        self._openLocalConfigFile(path)

    def openLocalExclude(self):
        path = self.repo.in_gitdir("info/exclude")
        self._openLocalConfigFile(path)

    def _openLocalConfigFile(self, fullPath: str):
        def createAndOpen():
            open(fullPath, "ab").close()
            ToolProcess.startTextEditor(self, fullPath)

        if not os.path.exists(fullPath):
            basename = os.path.basename(fullPath)
            askConfirmation(
                self,
                _("Open {0}", tquo(basename)),
                paragraphs(
                    _("File {0} does not exist.", bquo(fullPath)),
                    _("Do you want to create it?")),
                okButtonText=_("Create {0}", lquo(basename)),
                callback=createAndOpen)
        else:
            ToolProcess.startTextEditor(self, fullPath)

    def openTerminal(self):
        ToolProcess.startTerminal(self, self.workdir)

    def executeUserCommand(self, command: UserCommand):
        title = _("Run Command")

        try:
            compiledCommand = command.compile(self)
        except UserCommand.MultiTokenError as mte:
            errorText = (
                    _("The prerequisites for your command are not met:")
                    + f"<p><tt>{escape(command.command)}</tt></p>"
                    + toTightUL(f"{escape(str(error))} (<b>{escape(token)}</b>)"
                                for token, error in mte.tokenErrors.items()))
            showWarning(self, title, errorText)
            return

        def run():
            ToolProcess.startTerminal(self, self.workdir, compiledCommand)

        if command.alwaysConfirm or settings.prefs.confirmCommands:
            if not command.userTitle:
                question = _("Do you want to run this command in a terminal?")
            else:
                question = _("Do you want to run {0} in a terminal?").format(hquo(stripAccelerators(command.userTitle)))
            commandString = shlex.join(compiledCommand)
            question += f"<p><tt>{escape(commandString)}</tt></p>"
            askConfirmation(self, title, question, callback=run)
        else:
            run()

    # -------------------------------------------------------------------------
    # Entry point for generic "Find" command

    def dispatchSearchCommand(self, op: SearchBar.Op):
        searchBars = {
            self.sidebar: self.sidebar.searchBar,
            self.diffArea.dirtyFiles: self.diffArea.dirtyFiles.searchBar,
            self.diffArea.stagedFiles: self.diffArea.stagedFiles.searchBar,
            self.diffArea.committedFiles: self.diffArea.committedFiles.searchBar,
            self.diffArea.diffView: self.diffArea.diffView.searchBar,
        }

        # Find a sink to redirect search to
        focus = self.focusWidget()
        for sink, searchBar in searchBars.items():
            # Stop scanning if this sink or searchBar have focus
            if sink.isVisibleTo(self) and (focus is sink or focus is searchBar.lineEdit):
                break
        else:
            # Fall back to searching GraphView if nothing has focus
            searchBar = self.graphView.searchBar

        # Kick off search
        searchBar.popUp(op)

    # -------------------------------------------------------------------------

    def toggleHideRefPattern(self, refPattern: str, allButThis: bool = False):
        wasVisibleInGraph = self.graphView.isLocatorVisible(self.navLocator)

        assert refPattern.startswith("refs/")
        self.repoModel.toggleHideRefPattern(refPattern, allButThis)
        self.graphView.clFilter.updateHiddenCommits()

        # Hide/draw refboxes for commits that are shared by non-hidden refs
        self.graphView.viewport().update()

        # Re-jump to the locator if it just became visible/hidden.
        # This will display/hide the "hidden branch" banner.
        if wasVisibleInGraph != self.graphView.isLocatorVisible(self.navLocator):
            self.jump(self.navLocator)

    # -------------------------------------------------------------------------

    def refreshRepo(self):
        """Refresh the repo as soon as possible."""
        self.taskRunner.pendingEpilog.effects |= TaskEffects.DefaultRefresh
        self.onTaskRunnerReady()

    def onTaskRunnerReady(self):
        # Don't refresh if in background or task runner busy
        if not self.isVisible() or self.taskRunner.isBusy():
            return

        effects, jumpTo = self.taskRunner.consumePendingEffectsAndLocator()
        if effects:
            tasks.RefreshRepo.invoke(self, effects, jumpTo)
        elif jumpTo:
            tasks.Jump.invoke(self, jumpTo)

    def refreshWindowTitle(self):
        title = self.getTitle()
        inBrackets = ""
        repo = self.repo

        if repo.head_is_unborn:
            inBrackets = _("Unborn HEAD")
        elif repo.head_is_detached:
            oid = repo.head_commit_id
            inBrackets = f'{_("Detached HEAD")} @ {shortHash(oid)}'
        else:
            with suppress(GitError):
                inBrackets = repo.head_branch_shorthand

        if inBrackets:
            title = f"{title} [{inBrackets}]"

        self.setWindowTitle(title)

    def refreshBanner(self):
        """ Refresh state banner (merging, cherrypicking, reverting, etc.) """
        repo = self.repo

        rstate = repo.state() if repo else RepositoryState.NONE

        bannerTitle = trtables.enum(rstate) if rstate != RepositoryState.NONE else ""
        bannerText = ""
        bannerHeeded = False
        bannerAction = ""
        bannerCallback = None

        def abortMerge():
            tasks.AbortMerge.invoke(self)

        if rstate == RepositoryState.MERGE:
            mergingWhat = ""
            with suppress(IndexError, KeyError):
                mergehead = self.repoModel.mergeheads[0]
                mergingWhat = shortHash(mergehead)  # Take commit hash first in case refsAt raises KeyError
                mergingWhat = self.repoModel.refsAt[mergehead][0]
                mergingWhat = RefPrefix.split(mergingWhat)[1]
            bannerTitle = _("Merging {0}", bquo(mergingWhat))

            if not repo.any_conflicts:
                bannerText += _("All conflicts fixed. Commit to conclude.")
                bannerHeeded = True
            else:
                bannerText += _("Conflicts need fixing.")

            bannerAction = englishTitleCase(_("Abort merge"))
            bannerCallback = abortMerge

        elif rstate == RepositoryState.CHERRYPICK:
            if not repo.any_conflicts:
                bannerText += _("Commit to conclude the cherry-pick.")
                bannerHeeded = True
            else:
                bannerText += _("Conflicts need fixing.")

            bannerAction = englishTitleCase(_("Abort cherry-pick"))
            bannerCallback = abortMerge

        elif rstate == RepositoryState.REVERT:
            if not repo.any_conflicts:
                bannerText += _("Commit to conclude the revert.")
                bannerHeeded = True
            else:
                bannerText += _("Conflicts need fixing.")

            bannerAction = englishTitleCase(_("Abort revert"))
            bannerCallback = abortMerge

        elif rstate == RepositoryState.NONE:
            if repo.any_conflicts:
                bannerTitle = _("Conflicts")
                bannerText = _("Fix the conflicts among the uncommitted changes.")
                bannerAction = englishTitleCase(_("Reset index"))
                bannerCallback = abortMerge

        else:
            bannerTitle = _("Warning")
            bannerText = _(
                "The repo is currently in state {state}, which {app} doesn’t support yet. "
                "Use <code>git</code> on the command line to continue.",
                app=qAppName(), state=bquo(trtables.enum(rstate)))

        with DisableWidgetUpdatesContext(self.sideSplitter):
            if bannerText or bannerTitle:
                self.mergeBanner.popUp(bannerTitle, bannerText, heeded=bannerHeeded, canDismiss=False)
                if bannerAction:
                    self.mergeBanner.addButton(bannerAction, bannerCallback)
            else:
                self.mergeBanner.setVisible(False)

    def refreshNumUncommittedChanges(self):
        self.sidebar.repaintUncommittedChanges()
        self.graphView.repaintCommit(UC_FAKEID)

    # -------------------------------------------------------------------------

    def selectRef(self, refName: str):
        oid = self.repo.commit_id_from_refname(refName)
        self.jump(NavLocator(NavContext.COMMITTED, commit=oid))

    # -------------------------------------------------------------------------

    def onRepoTaskProgress(self, progressText: str, withSpinner: bool = False):
        if withSpinner:
            self.busyMessage.emit(progressText)
        elif progressText:
            self.statusMessage.emit(progressText)
        else:
            self.clearStatus.emit()

        if not withSpinner:
            self.busyCursorDelayer.stop()
            self.setCursor(Qt.CursorShape.ArrowCursor)
        elif not self.busyCursorDelayer.isActive():
            self.busyCursorDelayer.start()

    def onBusyCursorDelayerTimeout(self):
        self.setCursor(Qt.CursorShape.BusyCursor)

    def onRepoGone(self):
        message = _("Repository folder went missing:") + "\n" + escamp(self.workdir)
        self.replaceWithStub(message=message)

    def onAutoFetchTimerTimeout(self):
        if not settings.prefs.autoFetch or not self.isVisible() or self.taskRunner.isBusy():
            return

        # Check if it's time to auto-fetch.
        now = QDateTime.currentSecsSinceEpoch()
        interval = max(1, settings.prefs.autoFetchMinutes) * 60
        if now - self.lastAutoFetchTime > interval:
            AutoFetchRemotes.invoke(self)
            self.lastAutoFetchTime = now

    # -------------------------------------------------------------------------

    def processInternalLink(self, url: QUrl | str):
        if not isinstance(url, QUrl):
            url = QUrl(url)

        if url.isLocalFile():
            locator = NavLocator()
            fragment = url.fragment()
            if fragment:
                with suppress(ValueError):
                    locator = NavLocator.inCommit(Oid(hex=fragment))

            self.openRepo.emit(url.toLocalFile(), locator)
            return

        if url.scheme() != APP_URL_SCHEME:
            warnings.warn(f"Unsupported scheme in internal link: {url.toDisplayString()}")
            return

        logger.info(f"Internal link: {url.toDisplayString()}")

        simplePath = url.path().removeprefix("/")
        kwargs = dict(QUrlQuery(url).queryItems(QUrl.ComponentFormattingOption.FullyDecoded))

        if url.authority() == NavLocator.URL_AUTHORITY:
            locator = NavLocator.parseUrl(url)
            self.jump(locator)
        elif url.authority() == "expandlog":
            # After loading, jump back to what is currently the last commit
            jumpTo = NavLocator.inCommit(self.repoModel.commitSequence[-1].id)
            # Reload the repo
            maxCommits = int(kwargs.get("n", self.repoModel.nextTruncationThreshold))
            self.replaceWithStub(jumpTo, maxCommits)
        elif url.authority() == "prefs":
            self.openPrefs.emit(simplePath)
        else:  # pragma: no cover
            warnings.warn(f"Unsupported authority in internal link: {url.toDisplayString()}")

    # -------------------------------------------------------------------------

    def contextMenuItems(self):
        return self.contextMenuItemsByProxy(self, lambda: self)

    @classmethod
    def contextMenuItemsByProxy(cls, invoker, proxy):
        return [
            TaskBook.action(invoker, tasks.NewCommit, accel="C"),
            TaskBook.action(invoker, tasks.AmendCommit, accel="A"),
            TaskBook.action(invoker, tasks.NewStash),

            ActionDef.SEPARATOR,

            TaskBook.action(invoker, tasks.NewBranchFromHead, accel="B"),
            TaskBook.action(invoker, tasks.FetchRemotes, accel="F"),
            TaskBook.action(invoker, tasks.PullBranch, accel="L"),
            TaskBook.action(invoker, tasks.PushBranch, accel="P"),

            TaskBook.action(invoker, tasks.NewRemote),

            ActionDef.SEPARATOR,

            TaskBook.action(invoker, tasks.RecallCommit),

            ActionDef.SEPARATOR,

            # TODO: Yech (invoker.window())
            *invoker.window().repolessActions(lambda: proxy().workdir),

            ActionDef.SEPARATOR,

            ActionDef(
                _("&Local Config Files"),
                submenu=[
                    ActionDef(".gitignore", lambda: proxy().openGitignore()),
                    ActionDef("config", lambda: proxy().openLocalConfig()),
                    ActionDef("exclude", lambda: proxy().openLocalExclude()),
                ]),

            TaskBook.action(invoker, tasks.EditRepoSettings),
        ]

    @CallbackAccumulator.deferredMethod(250)
    def scheduleFlushGpgVerificationQueue(self):
        if self.taskRunner.isBusy():
            # Thanks to the deferredMethod decorator, this will reschedule
            # the call (instead of recursing).
            logger.debug("Rescheduling VerifyGpgQueue...")
            self.scheduleFlushGpgVerificationQueue()
            return

        VerifyGpgQueue.invoke(self)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID, RepoModel
from gitfourchette.toolbox import *
from gitfourchette.toolbox.filterchangecontext import FilterChangeContext


class CommitLogFilter(QSortFilterProxyModel):
    repoModel: RepoModel
    shadowHiddenIds: set[Oid]

    def __init__(self, repoModel, parent):
        super().__init__(parent)
        self.repoModel = repoModel
        self.shadowHiddenIds = set()
        self.shadowPathspecFilterActive = False
        self.setDynamicSortFilter(True)

        self.updateHiddenCommits()  # prime hiddenIds
        self.updatePathspecFilter()

    @property
    def pathspecFilter(self):
        return self.repoModel.commitPathspecFilter

    @benchmark
    def updateHiddenCommits(self):
        hiddenIds = self.repoModel.hiddenCommits

        # Invalidating the filter can be costly, so avoid if possible
        if self.shadowHiddenIds == hiddenIds:
            return

        with FilterChangeContext(self):
            # Keep a copy so we can detect a change next time we're called
            self.shadowHiddenIds = set(hiddenIds)

    @benchmark
    def updatePathspecFilter(self):
        active = self.pathspecFilter.wantFilter()

        # Invalidating the filter can be costly, so avoid if possible
        if not active and active == self.shadowPathspecFilterActive:
            return

        with FilterChangeContext(self):
            self.shadowPathspecFilterActive = active

    def filterAcceptsRow(self, sourceRow: int, sourceParent: QModelIndex) -> bool:
        try:
            commit = self.repoModel.commitSequence[sourceRow]
        except IndexError:
            # Probably an extra special row
            return True

        if commit.id == UC_FAKEID:
            # Always ignore shadowHiddenIds for the workdir row (same as pre–file-search
            # behavior). UC_FAKEID can appear in hiddenCommits graph bookkeeping; it must
            # not hide the synthetic uncommitted row.
            if self.shadowPathspecFilterActive:
                return commit.id in self.pathspecFilter.matchingIds
            return True

        if commit.id in self.shadowHiddenIds:
            return False

        if self.shadowPathspecFilterActive:
            return commit.id in self.pathspecFilter.matchingIds

        return True



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *


class FilterChangeContext:
    """
    Context manager wrapper for QSortFilterProxyModel.beginFilterChange and
    endFilterChange. Backwards compatible with Qt versions 6.10 and older
    which lack these functions.
    """

    def __init__(self, model: QSortFilterProxyModel):
        self.model = model

    def __enter__(self):
        try:  # Qt 6.9+
            self.model.beginFilterChange()
        except AttributeError:  # pragma: no cover - TODO: Remove once we can drop compatibility with Qt <6.9
            pass  # Let __exit__ do the actual work

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:  # Qt 6.10+
            self.model.endFilterChange(QSortFilterProxyModel.Direction.Rows)
        except AttributeError:  # pragma: no cover - TODO: Remove once we can drop compatibility with Qt <6.10
            self.model.invalidateFilter()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
from pathlib import Path
from typing import Any

from gitfourchette import trtables
from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.exttools.toolpresets import ToolPresets
from gitfourchette.exttools.usercommandsyntaxhighlighter import UserCommandSyntaxHighlighter
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.settings import SHORT_DATE_PRESETS, prefs
from gitfourchette.syntax import ColorScheme, PygmentsPresets
from gitfourchette.themes import ThemeName, ThemeColors, ThemeAccent
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)

SAMPLE_SIGNATURE = Signature("Jean-Michel Tartempion", "jm.tarte@example.com", 0, 0)
SAMPLE_FILE_PATH = "spam/.ham/eggs/hello.c"

# This dict initially contains hardcoded language names, e.g. to disambiguate
# both Portugueses, to give U.S. English a shorter name, etc. For all other
# languages, the dict is fleshed out with QLocale.nativeLanguageName().
LANGUAGE_NAMES = {
    "en": "English",
    "es": "Español",
    "pt": "Português PT",
    "pt_BR": "Português BR",
}


def _boxWidget(layoutType: type[QVBoxLayout | QHBoxLayout], *controls) -> QWidget:
    w = QWidget()
    layout: QBoxLayout = layoutType(w)
    layout.setSpacing(0)
    layout.setContentsMargins(0, 0, 0, 0)
    for control in controls:
        layout.addWidget(control)
    return w


def vBoxWidget(*controls):
    return _boxWidget(QVBoxLayout, *controls)


def hBoxWidget(*controls):
    return _boxWidget(QHBoxLayout, *controls)


def makeshiftSpacer(height=1):
    spacer = QWidget()
    spacer.setEnabled(False)
    spacer.setFixedSize(1, height)
    return spacer


def availableLocaleCodes() -> list[str]:
    """
    Returns gettext-compatible locale codes for which we have translation files.
    Note: English NOT included!
    """
    return [f.removesuffix(".mo") for f in QDir("assets:lang", "*.mo").entryList()]


def localeCodeToLanguageName(code: str) -> str:
    try:
        name = LANGUAGE_NAMES[code]
    except KeyError:
        # Cache native language name
        name = QLocale(code).nativeLanguageName()
        name = name or f"???{code}???"  # Fallback if language code not recognized by Qt
        name = name[0].upper() + name[1:]  # Many languages don't capitalize their name
        LANGUAGE_NAMES[code] = name

    return name


class PrefsDialog(QDialog):
    lastCategory = 0

    CategoryPrefix = "_category_"
    SpacerPrefix = "_spacer"
    LabelPrefix = "_label_"
    ControlQObjectNamePrefix = "prefctl_"
    LocCategoryHeaderSuffix = "_HEADER"
    LocSettingHelpSuffix = "_help"

    @benchmark
    def __init__(self, parent: QWidget, focusOn: str = ""):
        super().__init__(parent)

        self.setObjectName("PrefsDialog")
        self.setWindowTitle(_("{app} Settings", app=qAppName()))

        self.prefDiff: dict[str, Any] = {}
        "Delta to on-disk preferences."

        self.categoryKeys: list[str] = []

        self.categoryList = QListWidget()
        self.categoryList.setWordWrap(True)
        self.categoryList.setUniformItemSizes(True)
        self.categoryList.setMinimumWidth(200)
        self.categoryList.setMaximumWidth(200)
        self.categoryList.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.categoryList.setTextElideMode(Qt.TextElideMode.ElideRight)
        self.categoryList.currentRowChanged.connect(self.onCategoryChanged)
        self.categoryList.setIconSize(QSize(24, 24))

        self.categoryLabel = QLabel("CATEGORY")
        tweakWidgetFont(self.categoryLabel, 130)

        self.stackedWidget = QStackedWidget()

        buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Help)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        self.guideButton = buttonBox.button(QDialogButtonBox.StandardButton.Help)
        self.guideButton.setCheckable(True)
        self.guideButton.clicked.connect(self.toggleGuideBrowser)

        self.guideBrowser = QTextBrowser(self)
        self.guideBrowser.setMinimumWidth(400)
        self.guideBrowser.setOpenExternalLinks(True)
        self.guideBrowser.setVisible(False)
        tweakWidgetFont(self.guideBrowser, 90)

        layout = QGridLayout(self)
        layout.addWidget(self.categoryList,     0, 0, 4, 1)
        layout.addWidget(self.categoryLabel,    0, 1)
        layout.addWidget(QFaintSeparator(),     1, 1)
        layout.addWidget(self.stackedWidget,    2, 1)
        layout.addWidget(self.guideBrowser,     0, 2, 4, 1)
        self._fillControls(focusOn)
        layout.addWidget(buttonBox, 3, 1)  # Add buttonBox last so it comes last in tab order

        layout.setColumnStretch(0, 0)
        layout.setColumnStretch(1, 2)

        if not focusOn:
            # Restore last category
            self.setCategory(PrefsDialog.lastCategory)
            buttonBox.button(QDialogButtonBox.StandardButton.Ok).setFocus()
        else:
            # Save this category if we close the dialog without changing tabs
            PrefsDialog.lastCategory = self.stackedWidget.currentIndex()

        self.setModal(True)

    def _fillControls(self, focusOn: str):
        skipKeys = self.getHiddenSettingKeys()
        form: QFormLayout | None = None

        for key in prefs.__dict__:
            # New category tab
            if key.startswith(self.CategoryPrefix):
                category = key.removeprefix(self.CategoryPrefix)
                # Stop past the hidden category
                if category == "hidden":
                    break
                form = self._newCategoryForm(category)
                continue

            assert form is not None

            # Spacer
            if key.startswith(self.SpacerPrefix):
                form.addRow(makeshiftSpacer())
                continue

            # Label
            if key.startswith(self.LabelPrefix):
                labelKey = key.removeprefix(self.LabelPrefix)
                labelText = trtables.prefKey(labelKey)
                label = QLabel(labelText)
                label.setEnabled(False)
                tweakWidgetFont(label, bold=True)
                if form.count():  # add a spacer before the label
                    form.addRow(makeshiftSpacer())
                form.addRow(label)
                continue

            # Skip hidden settings
            if key in skipKeys or key.startswith("_"):
                continue

            # Add the control to the form layout, with a leading caption if any
            control, label, field = self._newRow(key)
            if label is not None:
                form.addRow(label, field)
            else:
                form.addRow(field)

            # If the current key matches the setting we want to focus on,
            # bring this tab to the foreground
            if focusOn == key:
                categoryIndex = self.stackedWidget.indexOf(form.parentWidget())
                self.setCategory(categoryIndex)
                control.setFocus()

    def _newCategoryForm(self, category: str) -> QFormLayout:
        formContainer = QWidget(self)

        form = QFormLayout(formContainer)
        form.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)

        categoryName = trtables.prefKey(category)
        categoryIcon = stockIcon(f"prefs-{category.lower()}")

        self.categoryKeys.append(category)
        self.stackedWidget.addWidget(formContainer)
        self.categoryList.addItem(QListWidgetItem(categoryIcon, categoryName))

        headerText = trtables.prefKeyNoDefault(category + self.LocCategoryHeaderSuffix)
        if headerText:
            headerText = headerText.format(app=qAppName())
            explainer = QLabel(headerText)
            explainer.setWordWrap(True)
            explainer.setTextFormat(Qt.TextFormat.RichText)
            tweakWidgetFont(explainer, 88)
            form.addRow(explainer)

        return form

    def _newRow(self, key: str) -> tuple[QWidget, QLabel | None, QWidget | QLayout]:
        """
        Build the widgets representing the given setting.
        Return tuple: main control widget, label (if any), field to be inserted
        into the QFormLayout.
        """

        # Get caption and suffix
        suffix = ""
        caption = trtables.prefKey(key)
        if "#" in caption:
            caption, suffix = caption.split("#")
            caption = caption.rstrip()
            suffix = suffix.lstrip()

        # Get the value of this setting
        prefValue = prefs.__dict__[key]

        # Make the actual control widget
        control = self.makeControlWidget(key, prefValue, caption)
        rowWidgets = [control]

        # Name the control so that unit tests can find it
        control.setObjectName(self.ControlQObjectNamePrefix + key)

        # Tack an extra QLabel to the end if there's a suffix
        if suffix:
            rowWidgets.append(QLabel(suffix))

        if key == "autoFetchMinutes":
            self.prependCheckBox(rowWidgets, "autoFetch", caption)

        # Any help text? Then make a help button for it & set tooltip text on the main control
        tip = trtables.prefKeyNoDefault(key + self.LocSettingHelpSuffix)
        if tip:
            tip = tip.format(app=qAppName())
            control.setToolTip(tip)
            hintButton = QHintButton(self, tip)
            hintButton.setMaximumHeight(2 + hintButton.fontMetrics().height())
            rowWidgets.append(hintButton)

        # Gather what to add to the form as a single item.
        # If we have more than a single widget to add to the form, lay them out in a row.
        formField: QWidget | QLayout
        if len(rowWidgets) == 1:
            formField = control
        else:
            rowLayout = QHBoxLayout()
            for w in rowWidgets:
                rowLayout.addWidget(w)
            if control.sizePolicy().horizontalPolicy() == QSizePolicy.Policy.Minimum:
                # Stick help button to right edge of non-expanding widget
                rowLayout.addStretch()
            formField = rowLayout

        # No caption, make field span entire row
        if not caption or isinstance(rowWidgets[0], QCheckBox):
            return control, None, formField

        # There's a leading caption, so add it as the label in the row
        caption += _(":")
        label = QLabel(caption)
        label.setBuddy(rowWidgets[0])
        if tip:
            label.setToolTip(tip)
        return control, label, formField

    def setCategory(self, row: int):
        self.categoryList.setCurrentRow(row)

    def onCategoryChanged(self, row: int):
        categoryKey = self.categoryKeys[row]
        categoryName = trtables.prefKey(categoryKey)
        categoryGuide = trtables.prefKeyNoDefault(f"{categoryKey}_guide")

        self.stackedWidget.setCurrentIndex(row)
        self.categoryLabel.setText(categoryName)

        self.toggleGuideBrowser(False)
        if categoryGuide:
            self.guideButton.setText(_("{0} Handy Reference").format(categoryName))
            self.guideButton.setVisible(True)
            self.guideBrowser.setHtml(categoryGuide)
        else:
            self.guideButton.setVisible(False)

        # Remember which tab we've last clicked on for next time we open the dialog
        PrefsDialog.lastCategory = row

    def toggleGuideBrowser(self, show: bool):
        if show == self.guideBrowser.isVisible():
            pass
        elif show:
            self._widthBeforeGuide = self.width()
            self.guideBrowser.show()
        else:
            self.guideBrowser.hide()
            QTimer.singleShot(0, lambda: self.resize(self._widthBeforeGuide, self.height()))
        self.guideButton.setChecked(show)

    def assign(self, k, v):
        if prefs.__dict__[k] == v:
            if k in self.prefDiff:
                del self.prefDiff[k]
        else:
            self.prefDiff[k] = v
        logger.debug(f"Assign {k} {v} ({type(v)})")

    def getMostRecentValue(self, k):
        if k in self.prefDiff:
            return self.prefDiff[k]
        elif k in prefs.__dict__:
            return prefs.__dict__[k]
        else:
            return None

    def getHiddenSettingKeys(self) -> set[str]:
        skipKeys = {
            "fontSize",  # bundled with "font"
        }

        # Prevent hiding menubar on macOS
        if MACOS:
            skipKeys.add("showMenuBar")

        # In frozen distributions, hide settings that depend on system Python
        # packages outside our sandbox.
        if APP_FREEZE_QT:
            skipKeys.add("forceQtApi")
            skipKeys.add("pygmentsPlugins")

        return skipKeys

    def makeControlWidget(self, key: str, value, caption: str) -> QWidget:
        valueType = type(value)

        if key == "language":
            return self.languageControl(key, value)
        elif key == "qtStyle":
            return self.qtStyleControl(key, value)
        elif key == "font":
            return self.fontControl(key)
        elif key == "shortTimeFormat":
            return self.dateFormatControl(key, value, SHORT_DATE_PRESETS)
        elif key == "pathDisplayStyle":
            return self.enumControl(key, value, valueType, previewCallback=lambda v: abbreviatePath(SAMPLE_FILE_PATH, v))
        elif key == "authorDisplayStyle":
            return self.enumControl(key, value, valueType, previewCallback=lambda v: abbreviatePerson(SAMPLE_SIGNATURE, v))
        elif key == "shortHashChars":
            return self.boundedIntControl(key, value, 4, 40)
        elif key == "maxRecentRepos":
            return self.boundedIntControl(key, value, 0, 50)
        elif key == "contextLines":  # staging/discarding individual lines is flaky with 0 context lines
            return self.boundedIntControl(key, value, 1, 32)
        elif key == "tabSpaces":
            return self.boundedIntControl(key, value, 1, 16)
        elif key == "autoFetchMinutes":
            return self.boundedIntControl(key, value, 1, 9999)
        elif key == "syntaxHighlighting":
            return self.syntaxHighlightingControl(key, value)
        elif key == "colorblind":
            return self.colorblindControl(key, value)
        elif key == "maxCommits":
            control = self.boundedIntControl(key, value, 0, 999_999_999, 1000)
            control.setSpecialValueText("\u221E")  # infinity
            return control
        elif key == "externalEditor":
            return self.strControlWithPresets(key, value, ToolPresets.Editors, leaveBlankHint=True)
        elif key == "externalDiff":
            return self.strControlWithPresets(
                key, value, ToolPresets.DiffTools,
                validate=lambda cmd: ToolCommands.checkCommand(cmd, "$L", "$R"))
        elif key == "externalMerge":
            return self.strControlWithPresets(
                key, value, ToolPresets.MergeTools,
                validate=lambda cmd: ToolCommands.checkCommand(cmd, "$L", "$R", "$B", "$M"))
        elif key == "terminal":
            return self.strControlWithPresets(
                key, value, ToolPresets.Terminals,
                validate=lambda cmd: ToolCommands.checkCommand(cmd, "$COMMAND"))
        elif key == "commands":
            return self.userCommandTextEditControl(key, value)
        elif key in ["largeFileThresholdKB", "imageFileThresholdKB", "maxTrashFileKB"]:
            control = self.boundedIntControl(key, value, 0, 999_999)
            control.setSpecialValueText("\u221E")  # infinity
            return control
        elif key == "gitPath":
            presets = {}
            builtInGit = ToolPresets.flatpakBuiltInGit()
            if builtInGit:
                presets[_("Built-in git (sandboxed)")] = builtInGit
            presets[_("Auto-detected system git")] = ToolPresets.defaultGit(hostOnly=True)
            return self.strControlWithPresets(key, value, presets)
        elif issubclass(valueType, enum.Enum):
            return self.enumControl(key, value, type(value))
        elif valueType is int:
            return self.intControl(key, value)
        elif valueType is bool:
            trueText = trtables.prefKeyNoDefault(key + "_true")
            falseText = trtables.prefKeyNoDefault(key + "_false")
            if trueText or falseText:
                return self.boolComboBoxControl(key, value, trueName=trueText, falseName=falseText)
            else:
                return self.boolCheckBoxControl(key, value, caption)
        else:
            raise NotImplementedError(f"Write pref widget for {key}")

    @benchmark
    def languageControl(self, prefKey: str, prefValue: str):
        defaultCaption = _p("system default language setting", "System default")
        control = QComboBox(self)
        control.addItem(defaultCaption, userData="")
        control.insertSeparator(1)

        localeRatios = {}
        for wipLine in Path(QFile("assets:lang/wip.txt").fileName()).read_text().strip().splitlines():
            code, ratio = wipLine.split(" ")
            localeRatios[code] = ratio
        localeRatios["en"] = "100"

        localeCodes = availableLocaleCodes()
        assert "en" not in localeCodes, "English shouldn't have an .mo file"
        localeCodes.append("en")

        localeNames = {code: localeCodeToLanguageName(code) for code in localeCodes}
        localeCodes.sort(key=lambda code: "0" if code == "en" else localeNames[code].casefold())

        for code in localeCodes:
            name = f"{localeNames[code]} ({localeRatios.get(code, '--')}%)"
            control.addItem(name, code)

        control.setCurrentIndex(control.findData(prefValue))
        control.activated.connect(lambda index: self.assign(prefKey, control.currentData(Qt.ItemDataRole.UserRole)))
        control.setMaxVisibleItems(20)

        return control

    def fontControl(self, prefKey: str):
        fontControl = FontPicker(self)

        familyKey = prefKey
        sizeKey = "fontSize"

        def assignFont(family: str, size: int):
            self.assign(familyKey, family)
            self.assign(sizeKey, size)

        fontControl.setCurrentFont(self.getMostRecentValue(familyKey), self.getMostRecentValue(sizeKey))
        fontControl.assign.connect(assignFont)
        return fontControl

    def strControlWithPresets(self, prefKey, prefValue, presets, leaveBlankHint=False, validate=None):
        control = QComboBoxWithPreview(self)
        control.setEditable(True)

        for k in presets:
            preview = presets[k]
            if not preview and leaveBlankHint:
                preview = "- " + _p("hint user to leave the field blank", "leave blank") + " -"
            control.addItemWithPreview(k, presets[k], preview)
            if prefValue == presets[k]:
                control.setCurrentIndex(control.count()-1)

        if leaveBlankHint:
            control.lineEdit().setPlaceholderText(_("Leave blank for system default."))

        control.setEditText(prefValue)
        control.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred))

        control.editTextChanged.connect(lambda text: self.assign(prefKey, text))

        if validate:
            validator = ValidatorMultiplexer(self)
            validator.connectInput(control.lineEdit(), validate, mustBeValid=False)
            validator.run()

        return control

    def userCommandTextEditControl(self, prefKey, prefValue):
        monoFont = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        fontMetrics = QFontMetricsF(monoFont)

        control = QPlainTextEdit(self)
        control.setFont(monoFont)
        control.setMinimumWidth(round(fontMetrics.horizontalAdvance("x" * 72)))
        control.setTabStopDistance(fontMetrics.horizontalAdvance(" " * 4))

        highlighter = UserCommandSyntaxHighlighter(control)
        highlighter.setDocument(control.document())

        control.setPlaceholderText(_(
            "# Enter custom terminal commands here.\n"
            "# You can then launch them from the {menu} menu.\n"
            "# Click {button} below for more information.",
            menu=tquo(stripAccelerators(_("&Commands"))),
            button=tquo(_("Handy Reference"))))

        control.setPlainText(prefValue)
        control.textChanged.connect(lambda: self.assign(prefKey, control.toPlainText()))
        return control

    def intControl(self, prefKey, prefValue):
        control = QLineEdit(str(prefValue), self)
        control.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        control.setValidator(QIntValidator())
        control.textEdited.connect(lambda v, k=prefKey: self.assign(k, int(v) if v else 0))
        return control

    def boundedIntControl(self, prefKey, prefValue, minValue, maxValue, step=1):
        control = QSpinBox(self)
        control.setMinimum(minValue)
        control.setMaximum(maxValue)
        control.setValue(prefValue)
        control.setSingleStep(step)
        control.setGroupSeparatorShown(True)
        control.setAlignment(Qt.AlignmentFlag.AlignRight)
        control.setStepType(QSpinBox.StepType.AdaptiveDecimalStepType)
        control.valueChanged.connect(lambda v, k=prefKey: self.assign(k, v))
        return control

    def boolComboBoxControl(self, prefKey: str, prefValue: bool, falseName: str, trueName: str) -> QComboBox:
        control = QComboBox(self)
        control.addItem(trueName)  # index 0 --> True
        control.addItem(falseName)  # index 1 --> False
        control.setCurrentIndex(int(not prefValue))
        control.activated.connect(lambda index: self.assign(prefKey, index == 0))
        return control

    def boolCheckBoxControl(self, prefKey: str, prefValue: bool, caption: str) -> QCheckBox:
        control = QCheckBox(caption, self)
        control.setChecked(prefValue)
        control.checkStateChanged.connect(lambda state, k=prefKey: self.assign(k, state == Qt.CheckState.Checked))
        return control

    def enumControl(self, prefKey, prefValue, enumType, previewCallback=None) -> QComboBox | QComboBoxWithPreview:
        control: QComboBox | QComboBoxWithPreview
        if previewCallback:
            control = QComboBoxWithPreview(self)
        else:
            control = QComboBox(self)

        for enumMember in enumType:
            # PySide6 demotes StrEnum to str when stored with QComboBox.setItemData().
            # Wrap the value in a tuple to preserve the type. (PyQt5 & PyQt6 do the right thing here)
            data = (enumMember,)
            name = trtables.enum(enumMember)

            if name == "":
                continue

            if previewCallback:
                control.addItemWithPreview(name, data, previewCallback(enumMember))  # type: ignore[attr-defined] # mypy not smart enough here
            else:
                control.addItem(name, data)
            if prefValue == enumMember:
                control.setCurrentIndex(control.count() - 1)

            if data == ("",):
                control.insertSeparator(control.count())

        control.activated.connect(lambda i: self.assign(prefKey, control.itemData(i)[0]))  # unpack the tuple!

        return control

    def qtStyleControl(self, prefKey, prefValue):
        currentStyleName = prefValue.split(",", 1)[0]
        control = QComboBox(self)
        variantPicker = self._customThemeVariantPickerControl(prefValue)

        separator = ("", "")
        defaultStyle = (_p("system default theme setting", "System default"), "")
        nativeStyles = [(name, name) for name in QStyleFactory.keys()]  # noqa: SIM118
        customStyles = [(trtables.enum(theme), str(theme)) for theme in ThemeName]
        nativeStyles.sort()
        customStyles.sort()

        entries = [defaultStyle, *customStyles, separator, *nativeStyles]
        for caption, styleName in entries:
            if not caption and not styleName:
                control.insertSeparator(control.count())
            else:
                control.addItem(caption, userData=styleName)
                if styleName == currentStyleName:
                    control.setCurrentIndex(control.count() - 1)

        def onPickStyle():
            i = control.currentIndex()
            newValue = control.itemData(i, Qt.ItemDataRole.UserRole)
            if newValue in ThemeName:
                variantPicker.setVisible(True)
                accentIndex = variantPicker.currentIndex()
                accentName = variantPicker.itemData(accentIndex)
                newValue = accentName
            else:
                variantPicker.setVisible(False)
            self.assign(prefKey, newValue)

        control.activated.connect(onPickStyle)
        variantPicker.activated.connect(onPickStyle)
        variantPicker.setVisible(currentStyleName in ThemeName)

        group = QWidget(self)
        layout = QHBoxLayout(group)
        layout.setContentsMargins(QMargins())
        layout.addWidget(control)
        layout.addWidget(variantPicker)
        return group

    def _customThemeVariantPickerControl(self, prefValue: str) -> QComboBox:
        picker = QComboBox(self)
        picker.setIconSize(QSize(16, 16))
        enforceComboBoxMaxVisibleItems(picker, 32)

        picker.addItem(stockIcon("light-dark-toggle"), _("System colors"), str(ThemeName.BuiltIn))

        for dark in [False, True]:
            picker.insertSeparator(picker.count())

            themePrefix = ThemeName.BuiltIn + "," + ("dark" if dark else "light")
            theme = ThemeColors.resolveTheme(themePrefix)

            for accent in ThemeAccent:
                icon = stockIcon("theme-chip", f"white={theme.bg} black={theme.text} blue={accent}")
                caption = _("Dark {color}") if "dark" in themePrefix else _( "Light {color}")
                caption = caption.format(color=trtables.enum(accent))
                value = themePrefix + "," + accent
                picker.addItem(icon, caption, value)
                if value == prefValue:
                    picker.setCurrentIndex(picker.count() - 1)

        return picker

    def dateFormatControl(self, prefKey, prefValue, presets):
        currentDate = QDateTime.currentDateTime()
        sampleDate = QDateTime(QDate(currentDate.date().year(), 1, 30), QTime(9, 45))
        bogusTime = "Wednesday, December 99, 9999 99:99:99 AM"

        def genPreview(f):
            return QLocale().toString(sampleDate, f)

        def onEditTextChanged(text):
            preview.setText(genPreview(text))
            self.assign(prefKey, text)

        preview = QLabel(bogusTime)
        preview.setEnabled(False)
        preview.setMaximumWidth(preview.fontMetrics().horizontalAdvance(bogusTime))
        preview.setText(genPreview(prefValue))

        control = QComboBoxWithPreview(self)
        control.setEditable(True)
        for presetName, presetFormat in presets.items():
            control.addItemWithPreview(presetName, presetFormat, genPreview(presetFormat))
            if prefValue == presetFormat:
                control.setCurrentIndex(control.count()-1)
        control.setMinimumWidth(200)
        control.setEditText(prefValue)
        control.editTextChanged.connect(onEditTextChanged)

        return vBoxWidget(control, preview)

    @benchmark
    def syntaxHighlightingControl(self, prefKey, prefValue):
        autoCaption = _p("syntax highlighting", "Automatic ({name})", name=PygmentsPresets.Dark if isDarkTheme() else PygmentsPresets.Light)
        offCaption = _p("syntax highlighting", "Off")

        control = QComboBox(self)
        control.setIconSize(QSize(16, 16))  # Required if enforceComboBoxMaxVisibleItems kicks in
        control.addItem(stockIcon("light-dark-toggle"), autoCaption, userData=PygmentsPresets.Automatic)
        control.addItem(stockIcon("SP_BrowserStop"), offCaption, userData=PygmentsPresets.Off)
        control.insertSeparator(control.count())

        previousStyleName = ""
        for styleName, chipColors in ColorScheme.stylePreviews(prefs.pygmentsPlugins).items():
            # Insert a separator between light and dark themes, i.e. when sorting resets
            if styleName < previousStyleName:
                control.insertSeparator(control.count())
            previousStyleName = styleName
            # Little icon to preview the colors in this style
            chip = stockIcon("colorscheme-chip", chipColors)
            control.addItem(chip, styleName, userData=styleName)

        index = control.findData(prefValue)
        control.setCurrentIndex(index)

        def onPickStyle(index):
            pickedStyleName = control.itemData(index, Qt.ItemDataRole.UserRole)
            self.assign(prefKey, pickedStyleName)

        control.activated.connect(onPickStyle)

        control.setMaxVisibleItems(30)
        enforceComboBoxMaxVisibleItems(control)  # Prevent Fusion from creating a giant popup

        return control

    def colorblindControl(self, prefKey, prefValue):
        control = QComboBox(self)
        control.addItem(stockIcon("linebg-chip-redgreen"), _("Red and green"), userData=False)
        control.addItem(stockIcon("linebg-chip-colorblind"), _("Colorblind-friendly"), userData=True)

        index = control.findData(prefValue)
        control.setCurrentIndex(index)

        def onPickStyle(index):
            pickedStyleName = control.itemData(index, Qt.ItemDataRole.UserRole)
            self.assign(prefKey, pickedStyleName)

        control.activated.connect(onPickStyle)
        return control

    def prependCheckBox(self, rowWidgets: list[QWidget], booleanKey: str, caption: str) -> QCheckBox:
        """
        Add a QCheckBox on the same row as an existing widget,
        controlling whether another pref key is enabled or disabled.
        """

        def enableRowWidgets(enable: bool):
            for w in rowWidgets:
                if not isinstance(w, QCheckBox):
                    w.setEnabled(enable)

        booleanValue = prefs.__dict__[booleanKey]

        checkBox = self.boolCheckBoxControl(booleanKey, booleanValue, caption)
        checkBox.setObjectName(f"prefctl_{booleanKey}")  # Name the control so that unit tests can find it
        checkBox.checkStateChanged.connect(lambda state: enableRowWidgets(state == Qt.CheckState.Checked))

        rowWidgets.insert(0, checkBox)
        QWidget.setTabOrder(rowWidgets[0], rowWidgets[1])

        enableRowWidgets(booleanValue)  # Prime enabled/disabled state
        return checkBox



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations
from contextlib import suppress
from typing import ClassVar

import pygments.styles
import pygments.token

from gitfourchette.qt import *
from gitfourchette.toolbox.qtutils import isDarkTheme
from gitfourchette.toolbox.benchmark import benchmark

Token = pygments.token.Token


class PygmentsPresets:
    Automatic = ""
    Off = "off"
    Light = "stata-light"
    Dark = "stata-dark"


class ColorScheme:
    fallbackScheme: ClassVar[ColorScheme] = None
    _cachedScheme: ClassVar[ColorScheme] = None
    _cachedPreviews: ClassVar[dict[str, str]] = {}

    name: str
    backgroundColor: QColor
    foregroundColor: QColor
    scheme: dict
    highContrastScheme: dict

    def __init__(self):
        self.name = ""
        self.scheme = {}
        self.highContrastScheme = {}

        palette = QApplication.palette()
        self.backgroundColor = palette.color(QPalette.ColorRole.Base)
        self.foregroundColor = palette.color(QPalette.ColorRole.Text)

    def __bool__(self):
        return bool(self.scheme)

    def isDark(self):
        return self.backgroundColor.lightnessF() < .5

    def primeHighContrastVersion(self):
        """
        Prepare a high-contrast alternative where colors pop against red/green backgrounds
        """

        if self.highContrastScheme:
            return

        isDarkBackground = self.isDark()

        for tokenType, lowContrastCharFormat in self.scheme.items():
            charFormat = QTextCharFormat(lowContrastCharFormat)

            fgColor = charFormat.foreground().color()
            if isDarkBackground:
                fgColor = fgColor.lighter(150)
            else:
                fgColor = fgColor.darker(130)

            charFormat.setForeground(fgColor)
            charFormat.clearBackground()

            self.highContrastScheme[tokenType] = charFormat

    def basicQss(self, widget: QWidget):
        if not bool(self):
            return "/* NO PYGMENTS STYLE */"

        bg = self.backgroundColor.name()
        fg = self.foregroundColor.name()
        return f"{type(widget).__name__} {{ background-color: {bg}; color: {fg}; }}"

    @classmethod
    def resolve(cls, name: str) -> ColorScheme:
        # Resolve style alias
        if name == PygmentsPresets.Automatic:
            name = PygmentsPresets.Dark if isDarkTheme() else PygmentsPresets.Light
        if name == PygmentsPresets.Off:
            return cls.fallbackScheme

        if cls._cachedScheme.name == name:
            return cls._cachedScheme

        style = pygments.styles.get_style_by_name(name)

        scheme = ColorScheme()
        scheme.name = name
        scheme.backgroundColor = QColor(style.background_color)

        # Unpack style colors
        # (Intentionally skipping 'bgcolor' to prevent confusion with red/green backgrounds)
        for tokenType, styleForToken in style:
            charFormat = QTextCharFormat()
            if styleForToken['color']:
                assert not styleForToken['color'].startswith('#')
                color = QColor('#' + styleForToken['color'])
                charFormat.setForeground(color)
            if styleForToken['bold']:
                charFormat.setFontWeight(QFont.Weight.Bold)
            if styleForToken['italic']:
                charFormat.setFontItalic(True)
            if styleForToken['underline']:
                charFormat.setFontUnderline(True)
            scheme.scheme[tokenType] = charFormat

        with suppress(KeyError):
            scheme.foregroundColor = scheme.scheme[Token.Text].foreground().color()

        # Set consistent whitespace color
        wsColor = QColor(scheme.foregroundColor)
        wsColor.setAlphaF(.18)
        wsFormat = QTextCharFormat()
        wsFormat.setForeground(wsColor)
        scheme.scheme[Token.Whitespace] = wsFormat

        cls._cachedScheme = scheme
        return scheme

    @property
    def whitespaceFormat(self) -> QTextCharFormat:
        assert self.scheme
        return self.scheme[Token.Whitespace]

    @classmethod
    @benchmark
    def stylePreviews(cls, withPlugins: bool) -> dict[str, str]:
        if cls._cachedPreviews:
            return cls._cachedPreviews

        # pygments.styles.STYLES appeared in Pygments 2.17,
        # but we're maintaining backwards compatibility with old Pygments versions for now.
        if not hasattr(pygments.styles, 'STYLES'):  # pragma: no cover
            withPlugins = True

        if withPlugins:
            allStyles = pygments.styles.get_all_styles()
        else:
            allStyles = (styleName for _dummy1, styleName, _dummy2 in pygments.styles.STYLES.values())

        def extractColor(style, *tokenTypes):
            for t in tokenTypes:
                with suppress(TypeError):
                    return QColor('#' + style.style_for_token(t)['color'])
            return QColor(Qt.GlobalColor.black)

        primary: dict[str, str] = {}
        secondary: dict[str, str] = {}

        for styleName in sorted(allStyles):
            style = pygments.styles.get_style_by_name(styleName)
            bgColor = QColor(style.background_color)
            accent1 = extractColor(style, Token.Name.Class, Token.Text)
            accent2 = extractColor(style, Token.Name.Function, Token.Operator)
            accent3 = extractColor(style, Token.Keyword, Token.Comment)

            # Little icon to preview the colors in this style (colorscheme-chip.svg)
            chipColors = f"black={bgColor.name()} white={accent1.name()} red={accent2.name()} blue={accent3.name()}"

            # Sort light and dark themes in separate tables
            dark = bgColor.lightnessF() < .5
            table = secondary if dark else primary
            table[styleName] = chipColors

        primary.update(secondary)
        cls._cachedPreviews = primary
        return primary

    @classmethod
    def refreshFallbackScheme(cls):
        fallbackScheme = cls()
        cls.fallbackScheme = fallbackScheme
        cls._cachedScheme = fallbackScheme

    @classmethod
    def fillInFallback(cls, scheme: dict, tokenType) -> QTextCharFormat:
        assert Token in scheme
        assert tokenType not in scheme

        # Split tokenType from most specific to most generic type, i.e. tokenType first, Token last
        split = tokenType.split()
        split.reverse()
        assert split[0] == tokenType
        assert split[-1] == Token

        fallbackType = next(t for t in split if t in scheme)
        fallbackFormat: QTextCharFormat = scheme[fallbackType]

        for t in split:
            if t not in scheme:
                scheme[t] = fallbackFormat

        return fallbackFormat


ColorScheme.refreshFallbackScheme()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from pathlib import Path

from gitfourchette.qt import *
from gitfourchette.toolbox.qtutils import mixColors

_Gray = "gray"
"""
Name of the placeholder color in the icons' SVG markup.
This color will be replaced with IconColors.mainColor.
"""


class RecolorSvgIconEngine(QIconEngine):
    """
    QIcon renderer that adjusts the colors in an SVG icon. The colors are chosen
    based on the global color palette, active/disabled/selected icon modes, and
    an optional color remapping table.
    """

    class IconColors:
        background = QColor(0xFFFFFF)
        foreground = QColor(0x000000)
        highlight = QColor(0x00FFFF)
        mainColor = QColor(0xFF00FF if APP_DEBUG else 0x808080)
        preferDarkVariants = False
        systemFont = "sans-serif"
        initialized = False

        @classmethod
        def refresh(cls):
            palette = QApplication.palette()
            cls.background = palette.color(QPalette.ColorRole.Window)
            cls.foreground = palette.color(QPalette.ColorRole.WindowText)
            cls.highlight = palette.color(QPalette.ColorRole.HighlightedText)
            cls.mainColor = mixColors(cls.background, cls.foreground, .58)
            cls.preferDarkVariants = cls.background.lightness() < cls.foreground.lightness()
            cls.systemFont = QFontDatabase.systemFont(QFontDatabase.SystemFont.GeneralFont).family()
            cls.initialized = True

    def __init__(self, iconPathStr: str, colorTable: str = ""):
        super().__init__()

        # Initialize color scheme
        if not RecolorSvgIconEngine.IconColors.initialized:
            RecolorSvgIconEngine.IconColors.refresh()

        # Read in SVG data
        iconPath = Path(iconPathStr)
        assert iconPath.name.endswith(".svg")
        svg = iconPath.read_text("utf-8").strip()

        # Inject system font
        svg = svg.replace("sans-serif", RecolorSvgIconEngine.IconColors.systemFont)

        # Replace colors in hardcoded color table
        hardcodedColors = {}
        for entry in colorTable.split():
            original, replacement = entry.split("=", 1)
            hardcodedColors[original] = replacement
            svg = svg.replace(original, "{" + original + "}")
        svg = svg.format_map(hardcodedColors)

        # Inject format token for gray (unless colorTable overrides it)
        if _Gray not in hardcodedColors:
            svg = svg.replace(_Gray, "{" + _Gray + "}")

        # Inject format token for surrounding tag (used for opacity)
        beg = svg.index(">") + 1
        end = svg.rindex("</svg>")
        svg = svg[:beg] + "{prefix}" + svg[beg:end] + "{suffix}" + svg[end:]

        self.svg = svg
        self.informalIconName = iconPath.name  # for debugging
        self.basePixmapKey = hash(svg)
        self.renderers: dict[QIcon.Mode, QSvgRenderer] = {}
        self.referenceSize = QSize(0, 0)

        self.initVariants()

        from gitfourchette.application import GFApplication
        GFApplication.instance().restyle.connect(self.initVariants)

    def initVariants(self):
        IC = RecolorSvgIconEngine.IconColors
        self.renderers = {
            QIcon.Mode.Normal: self._recolor(IC.mainColor),
            QIcon.Mode.Disabled: self._recolor(IC.mainColor, opacity=.33),
            QIcon.Mode.Selected: self._recolor(IC.highlight),
            QIcon.Mode.SelectedInactive: self._recolor(IC.foreground),  # type: ignore[attr-defined]
        }
        self.referenceSize = self.renderers[QIcon.Mode.Normal].defaultSize()

    def iconName(self):
        return self.informalIconName

    def actualSize(self, size: QSize, mode: QIcon.Mode, state: QIcon.State):
        # Crispness hack: Tweak width/height so that centered X/Y coordinates
        # snap to integers. This fixes blurriness at 1x scaling when rendering,
        # for example, a 16x16 SVG into a 16x19 rectangle.
        rw, rh = self.referenceSize.width(), self.referenceSize.height()
        sw, sh = size.width(), size.height()
        if sw > rw and (sw ^ rw) & 1 != 0:
            sw -= 1
        if sh > rh and (sh ^ rh) & 1 != 0:
            sh -= 1
        return QSize(sw, sh)

    def paint(self, painter: QPainter, rect: QRect, mode: QIcon.Mode, state: QIcon.State):
        try:
            renderer = self.renderers[mode]
        except KeyError:
            renderer = self.renderers[QIcon.Mode.Normal]
        rectf = QRectF(rect)
        renderer.render(painter, rectf)

    def pixmap(self, size: QSize, mode: QIcon.Mode, state: QIcon.State):
        if not QT5:
            key = f"{self.basePixmapKey},{size.width()},{size.height()},{mode.value},{state.value}"
        else:  # pragma: no cover
            key = f"{self.basePixmapKey},{size.width()},{size.height()},{mode},{state}"
        pixmap = QPixmapCache.find(key)
        if pixmap is not None:
            return pixmap

        pixmap = QPixmap(size)
        pixmap.fill(Qt.GlobalColor.transparent)
        painter = QPainter(pixmap)
        rect = QRect(QPoint(0, 0), size)
        self.paint(painter, rect, mode, state)
        painter.end()

        QPixmapCache.insert(key, pixmap)
        return pixmap

    def _recolor(self, replaceGray: QColor, opacity: float = 1.0) -> QSvgRenderer:
        """
        Create a QSvgRenderer for a recolored variant of the base SVG icon.
        """

        tokens = {
            _Gray: replaceGray.name(),
            "prefix": "",
            "suffix": ""
        }

        if opacity != 1.0:
            tokens["prefix"] = f"<g opacity='{opacity}'>"
            tokens["suffix"] = "</g>"

        data = self.svg.format_map(tokens)
        blob = data.encode("utf-8")

        renderer = QSvgRenderer(blob)
        renderer.setAspectRatioMode(Qt.AspectRatioMode.KeepAspectRatio)
        return renderer



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from contextlib import suppress

from gitfourchette.qt import *


class QProcessConnection(QObject):
    """
    Emits processLost when a QProcess stops running or fails to start outright.
    """
    processLost = Signal()

    process: QProcess | None

    def __init__(self, parent):
        super().__init__(parent)
        self.process = None

    def __bool__(self):
        return self.process is not None

    def track(self, process: QProcess):
        assert process is not self.process, "reconnecting to same process"

        self.stopTracking()
        assert self.process is None

        self.process = process
        process.errorOccurred.connect(self.stopTracking)
        process.finished.connect(self.stopTracking)

    def stopTracking(self):
        process = self.process

        if process is not None:
            with suppress(TypeError, RuntimeError):
                process.finished.disconnect(self.stopTracking)
            with suppress(TypeError, RuntimeError):
                process.errorOccurred.disconnect(self.stopTracking)
            self.processLost.emit()

        self.process = None
        return process



# Form implementation generated from reading ui file 'deletetagdialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_DeleteTagDialog(object):
    def setupUi(self, DeleteTagDialog):
        DeleteTagDialog.setObjectName("DeleteTagDialog")
        DeleteTagDialog.setWindowModality(Qt.WindowModality.WindowModal)
        DeleteTagDialog.resize(405, 303)
        self.verticalLayout = QVBoxLayout(DeleteTagDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QLabel(parent=DeleteTagDialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pushCheckBox = QCheckBox(parent=DeleteTagDialog)
        self.pushCheckBox.setObjectName("pushCheckBox")
        self.horizontalLayout.addWidget(self.pushCheckBox)
        self.remoteComboBox = QComboBox(parent=DeleteTagDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteComboBox.sizePolicy().hasHeightForWidth())
        self.remoteComboBox.setSizePolicy(sizePolicy)
        self.remoteComboBox.setObjectName("remoteComboBox")
        self.horizontalLayout.addWidget(self.remoteComboBox)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.buttonBox = QDialogButtonBox(parent=DeleteTagDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(DeleteTagDialog)
        self.buttonBox.accepted.connect(DeleteTagDialog.accept)
        self.buttonBox.rejected.connect(DeleteTagDialog.reject)
        self.pushCheckBox.toggled['bool'].connect(self.remoteComboBox.setEnabled)

    def retranslateUi(self, DeleteTagDialog):
        DeleteTagDialog.setWindowTitle(_p("DeleteTagDialog", "Delete tag"))
        self.label.setText(_p("DeleteTagDialog", "Really delete tag {0}?"))
        self.pushCheckBox.setText(_p("DeleteTagDialog", "&Push deletion to:"))
